/* **********************************************************************
    Copyright 2003 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.rpi.cct.uwcal.calsvci;

import edu.rpi.cct.uwcal.resources.Resources;

import edu.washington.cac.calfacade.shared.AdminGroups;
import edu.washington.cac.calfacade.shared.CalendarVO;
import edu.washington.cac.calfacade.shared.CalFacadeDefs;
import edu.washington.cac.calfacade.shared.CalFacadeException;
import edu.washington.cac.calfacade.shared.Calintf;
import edu.washington.cac.calfacade.shared.CalintfFactory;
import edu.washington.cac.calfacade.shared.EventVO;
import edu.washington.cac.calfacade.shared.filter.FilterVO;
import edu.washington.cac.calfacade.shared.KeywordVO;
import edu.washington.cac.calfacade.shared.LocationVO;
import edu.washington.cac.calfacade.shared.MyCalendarVO;
import edu.washington.cac.calfacade.shared.RecurVO;
import edu.washington.cac.calfacade.shared.SponsorVO;
import edu.washington.cac.calfacade.shared.SynchInfoVO;
import edu.washington.cac.calfacade.shared.SynchStateVO;
import edu.washington.cac.calfacade.shared.UserAuth;
import edu.washington.cac.calfacade.shared.UserVO;

import java.sql.Timestamp;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.TreeSet;
import java.util.Vector;
import org.apache.log4j.Logger;

/** This is an implementation of the admin service interface to the calendar
 * suite.
 *
 * @author Mike Douglass       douglm@rpi.edu
 */
public class CalSvc implements CalSvcI {
  private CalSvcIPars pars;

  private boolean debug;

  /** Used to see if applications need to force a refresh
   */
  private long publicLastmod;

  private transient Calintf cali;

  private transient Logger log;

  private Resources res = new Resources();

  /** (Re)initialise the object for a particular use.
   *
   * @param pars        Defines the global parameters for the object
   * @return Calintf    object for interactions with back-end.
   */
  public void init(CalSvcIPars pars) throws CalFacadeException {
    this.pars = pars;
    this.debug = pars.getDebug();

    // Not sure why we were doing this - maybe just to ensure things are working
    //getCal().touch();
  }

  /** Switch to the given non-null user. The access rights are determined by
   * the previous call to init.
   *
   * @param val         String user id
   */
  public void setUser(String val) throws CalFacadeException {
    getCal().setUser(val);
  }

  /** Signal the start of a sequence of operations. These overlap transactions
   * in that there may be 0 to many transactions started and ended within an
   * open/close call and many open/close calls within a transaction.
   *
   * <p>The open close methods are mainly associated with web style requests
   * and open will usually be called early in the incoming request handling
   * and close will always be called on the way out allowing the interface an
   * opportunity to reacquire (on open) and release (on close) any resources
   * such as connections.
   */
  public void open() throws CalFacadeException {
    getCal().open();
  }

  /** Call on the way out after handling a request..
   */
  public void close() throws CalFacadeException {
    getCal().close();
  }

  /** Start a (possibly long-running) transaction. In the web environment
   * this might do nothing. The endTransaction method should in some way
   * check version numbers to detect concurrent updates and fail with an
   * exception.
   */
  public void beginTransaction() throws CalFacadeException {
    getCal().beginTransaction();
  }

  /** End a (possibly long-running) transaction. In the web environment
   * this should in some way check version numbers to detect concurrent updates
   * and fail with an exception.
   */
  public void endTransaction() throws CalFacadeException {
    getCal().endTransaction();
  }

  /** Call if there has been an error during an update process.
   */
  public void rollbackTransaction() throws CalFacadeException {
    getCal().rollbackTransaction();
  }

  /** Get the set of internationalized resources for this session
   */
   public Resources getResources() {
     return res;
   }

  /** Get a name uniquely.identifying this system. This should take the form <br/>
   *   name@host
   * <br/>where<ul>
   * <li>name identifies the particular calendar system at the site</li>
   * <li>host part identifies the domain of the site.</li>..
   * </ul>
   *
   * @return String    globally unique system identifier.
   */
  public String getSysid() throws CalFacadeException{
    return getCal().getSysid();
  }

  /** Returns a value object representing the current user.
   *
   * @return UserVO       representing the current user
   */
  public UserVO getUserVO() throws CalFacadeException {
    return getCal().getUserVO();
  }

  /** Get a UserAuth object which allows the application to determine what
   * special rights the user has.
   *
   * <p>Note that the returned object may require a one time initialisation
   * call to set it up correctly.
   * @see edu.washington.cac.calfacade.shared.UserAuth#initialise(String, int)
   * @see edu.washington.cac.calfacade.shared.UserAuth#initialise(String, Object)
   *
   * @param user         String account name
   * @param par          parameter object
   * @return UserAuth    implementation.
   */
  public UserAuth getUserAuth(String user, Object par) throws CalFacadeException {
    UserAuth ua = getCal().getUserAuth();
    ua.initialise(user, getCal(), par);

    return ua;
  }

  /** Get an initialised UserAuth object
   *
   * @return UserAuth    implementation.
   */
  public UserAuth getUserAuth() throws CalFacadeException {
    return getCal().getUserAuth();
  }

  /** Get an AdminGroups object which allows the application to handle
   * administrative groups.
   *
   * @return AdminGroups    implementation.
   */
  public AdminGroups getAdminGroups() throws CalFacadeException {
    return getCal().getAdminGroups();
  }

  /** Refresh the users events cache
   */
  public void refreshEvents() throws CalFacadeException {
    getCal().refreshEvents();
  }

  /** Return true if we believe a refresh is required. After a return of
   * true this method willreturn false until somebody updates the db.
   *
   * @return boolean    time to refresh
   */
  public boolean refreshNeeded() throws CalFacadeException {
    /* See if the events were updated */
    long lastmod;

    lastmod = getCal().getPublicLastmod();

    if (lastmod == publicLastmod) {
      return false;
    }

    publicLastmod = lastmod;
    return true;
  }

  /* ====================================================================
   *                   Calendars and search
   * ================================================================== */

  /** Return a list of calendar we can present to the client as a list of
   * available calendars.
   *
   * @return Collection   of CalendarVO
   */
  public Collection getCalendars() throws CalFacadeException {
    return getCal().getCalendars();
  }

  /** Set a search filter using the suppplied search string
   *
   * @param val    String search parameters
   */
  public void setSearch(String val) throws CalFacadeException {
    getCal().setSearch(val);
  }

  /** Return the current search string
   *
   * @return  String     search parameters
   */
  public String getSearch() throws CalFacadeException {
    return getCal().getSearch();
  }

  /** Set the calendar we are interested in. This is represented by the name
   * of a calendar.
   *
   * @param  val     int id of calendar - 0 for none
   * @return boolean true for OK, false for unknown calendar
   */
  public boolean setCalendar(int val) throws CalFacadeException {
    return getCal().setCalendar(val);
  }

  /** Get the name of the current calendar we are displaying
   *
   * @return String    name of calendar or null for all events
   */
  public String getCalendarTitle() throws CalFacadeException {
    return getCal().getCalendarTitle();
  }

  /** This method is called to add a calendar. No consistency checks are
   * carried out at the moment.
   *
   * @param val  String values for db
   */
  public void addCalendarDef(String val) throws CalFacadeException {
    getCal().addCalendarDef(val);
  }

  /** Add a filter to the database. All references must be set.
   */
  public void addFilter(FilterVO val) throws CalFacadeException {
    getCal().addFilter(val);
  }

  /** Update a filter
   *
   * @param  filter           FilterVO object to upate
   * @exception CalFacadeException If there's a problem
   */
  public void updateFilter(FilterVO val) throws CalFacadeException {
    getCal().updateFilter(val);
  }

  /* * Set a filter which returns events starting after a given Timestamp.
   *
   * @param val   Timestamp object.
   * /
  public void setLastmodGECalendar(Timestamp val) throws CalFacadeException {
    getCal().setLastmodGECalendar(val);
  }

  /* * Set a filter which returns events starting after a given date of the
   * form yyyymmdd.
   *
   * @param date   String of form yyyymmdd.
   * /
  public void setLastmodGEDateCalendar(String date) throws CalFacadeException {
    getCal().setLastmodGEDateCalendar(date);
  }

  /* * Set a filter which returns events starting after a given date and time
   * of the form yyyymmddhhmm.
   *
   * @param dateTime   String of form yyyymmddhhmm.
   * /
  public void setLastmodGEDateTimeCalendar(String dateTime)
      throws CalFacadeException {
    getCal().setLastmodGEDateTimeCalendar(dateTime);
  }*/

  /** ===================================================================
   *                   Subscriptions
   *  =================================================================== */

  /** Ensure that the users subscriptions match a set of calendars
   *
   * @param crs      TreeSet of Integer calendar ids
   * @return int[]   Number of calendars subscribed to, and number of
   *                 calendars unsubscribed to
   */
  public int[] fixSubscriptions(TreeSet crs) throws CalFacadeException {
    return getCal().fixSubscriptions(crs);
  }

  /** Determine if user is subscribed to a calendar
   *
   * @param id       int calendar id
   * @return boolean true if subscribed to calendars
   */
  public boolean getSubscribed(int id) throws CalFacadeException {
    return getCal().getSubscribed(id);
  }

  /* ====================================================================
   *                   Keywords
   * ==================================================================== */

  /* * Return all keywords for this user
   *
   * @return KeywordVO[]  array of keywords value objects
   * /
  public KeywordVO[] getKeywords() throws CalFacadeException {
    return getCal().getKeywords();
  }*/

  /** Add a Keyword to the database. The id will be set in the parameter
   * object.
   *
   * @param val   KeywordVO object to be added
   */
  public void addKeyword(KeywordVO val) throws CalFacadeException {
    getCal().addKeyword(val);
  }

  /** Add a Keyword to the database and give it the assigned key.
   * Only for restores
   *
   * @param val   KeywordVO object to be added with key set
   */
  public void addKeywordWithKey(KeywordVO val) throws CalFacadeException {
    getCal().addKeywordWithKey(val);
  }

  /** Replace a keyword in the database.
   *
   * @param val   KeywordVO object to be replaced
   */
  public void replaceKeyword(KeywordVO val) throws CalFacadeException {
    getCal().replaceKeyword(val);
  }

  /** Delete a keyword. Also remove it from the current
   * user preferences (if any).
   *
   * @param val      KeywordVO keyword
   * @return int     0 if it was deleted.
   *                 1 if it didn't exist
   *                 2 if in use
   */
  public int deleteKeyword(KeywordVO val) throws CalFacadeException {
    /** Only allow delete if not in use
     */
    if (getCal().checkKeywordRefs(val)) {
      return 2;
    }

    /* Remove from preferences */
    getCal().getUserAuth().removeKeyword(null, val);

    /* Attempt to delete
     */
    if (getCal().deleteKeyword(val)) {
      return 0;
    }

    return 1; //doesn't exist
  }

  /* * Return a keyword with the given id
   *
   * @param id     int id of the keyword
   * @return KeywordVo object representing the keyword in question
   *                     null if it doesn't exist.
   * /
  public KeywordVO getKeyword(int id) throws CalFacadeException {
    return getCal().getKeyword(id);
  }*/

  /** Search the keywords owned by this user for a matching name
   *
   * @param k          KeywordVO object to match for
   * @return KeywordVO matching object or null if no match.
   */
  public KeywordVO findKeyword(KeywordVO k) throws CalFacadeException {
    Collection kk = getCal().getKeywords();

    if (kk == null) {
      return null;
    }

    String word = k.getWord();
    Iterator it = kk.iterator();

    while (it.hasNext()) {
      KeywordVO ki = (KeywordVO)it.next();

      if ((word.equals(ki.getWord()))) {
        return ki;
      }
    }

    return null;
  }

  /** Ensure a keyword exists. If it does sets the id in the object and returns
   * false (not added).
   * If not creates the keyword then sets the new id in the object and returns
   * true (added)..
   *
   * @param keyword   KeywordVO value object. If this object has the id set,
   *                  we assume the check was made previously.
   * @return boolean  true if we added the entry
   */
  public boolean ensureKeywordExists(KeywordVO keyword) throws CalFacadeException {
    if (keyword.getId() != 0) {
      // We've already checked it exists
      return false;
    }

    keyword.setIsPublic(pars.getPublicAdmin());
    keyword.setCreator(getUserVO());

    // Assume a new keyword - but see if we can find it.
    KeywordVO k = findKeyword(keyword);

    if (k != null) {
      keyword.setId(k.getId());
      return false;
    }

    // doesn't exist at this point, so we add it to db table

    getCal().addKeyword(keyword);

    return true;
  }

  /* ====================================================================
   *                   Public Keywords
   * ==================================================================== */

  /** Return public keywords excluding alerts.
   * (I don't think there's anything other than public keywords.
   *
   * @return Collection     of keywords
   */
  public Collection getPublicKeywords() throws CalFacadeException {
    Collection ks = getCal().getPublicKeywords();
    if (ks == null) {
      return null;
    }

    Vector v = new Vector();
    Iterator it = ks.iterator();

    while (it.hasNext()) {
      KeywordVO k = (KeywordVO)it.next();
      if (!getKeyAlert(k)) {
        v.addElement(k);
      }
    }

    return v;
  }

  /** Return alert keywords only.
   *
   * @return Collection    of alert keywords
   */
  public Collection getAlertKeywords() throws CalFacadeException {
    Collection ks = getCal().getPublicKeywords();
    if (ks == null) {
      return null;
    }

    Vector v = new Vector();
    Iterator it = ks.iterator();

    while (it.hasNext()) {
      KeywordVO k = (KeywordVO)it.next();
      if (getKeyAlert(k)) {
        v.addElement(k);
      }
    }

    return v;
  }

  /** Return public keywords including alerts.
   *
   * @return Collection     of keywords
   */
  public Collection getAllPublicKeywords() throws CalFacadeException {
    return getCal().getPublicKeywords();
  }

  /** Return given public keyword
   *
   * @return keywordVO    keyword or null if it doesn't exist
   */
  public KeywordVO getPublicKeyword(int id) throws CalFacadeException {
    return getCal().getPublicKeyword(id);
  }

  /* ====================================================================
   *                   Keyword attributes
   * ==================================================================== */

  /** Set keyword alert on/off
   *
   * @param KeywordVO  keyword object
   * @param val        boolean turn it on/off
   */
  public void setKeyAlert(KeywordVO k, boolean val) throws CalFacadeException {
    boolean on = k.matchesAttr(CalFacadeDefs.catClassAttrName,
                               CalFacadeDefs.catClassAlerts);
    if (val && !on) {
      addKeywordAttr(k, CalFacadeDefs.catClassAttrName,
                        CalFacadeDefs.catClassAlerts);
    } else if (!val && on) {
      removeKeywordAttr(k, CalFacadeDefs.catClassAttrName,
                           CalFacadeDefs.catClassAlerts);
    }
  }

  /** Is given keyword an alert?
   *
   * @param KeywordVO  keyword object
   * @return boolean   true if this is an alert keyword
   */
  public boolean getKeyAlert(KeywordVO k) throws CalFacadeException {
    return k.matchesAttr(CalFacadeDefs.catClassAttrName,
                         CalFacadeDefs.catClassAlerts);
  }

  /* ====================================================================
   *                   Locations
   * ==================================================================== */

  /** Return all locations for this user. Returns an empty collection for
   * no locations. If this is a public admin instance, this will return all
   * public locations owned by the current user. Otherwise it returns all
   * personal locations owned by the user.
   *
   * @return Collection        of location value objects
   */
  public Collection getLocations() throws CalFacadeException {
    return getCal().getLocations();
  }

  /** Add a location to the database. The id will be set in the parameter
   * object.
   *
   * @param val   LocationVO object to be added
   */
  public void addLocation(LocationVO val) throws CalFacadeException {
    getCal().addLocation(val);
  }

  /** Replace a location in the database.
   *
   * @param val   LocationVO object to be replaced
   */
  public void replaceLocation(LocationVO val) throws CalFacadeException {
    getCal().replaceLocation(val);
  }

  /** Return a location with the given id
   *
   * @param id     int id of the location
   * @return LocationVo object representing the location in question
   *                     null if it doesn't exist.
   */
  public LocationVO getLocation(int id) throws CalFacadeException {
    return getCal().getLocation(id);
  }

  /** Delete a given location
   *
   * @param val      LocationVO object to be deleted
   * @return int     0 if it was deleted.
   *                 1 if it didn't exist
   *                 2 if in use
   */
  public int deleteLocation(LocationVO val) throws CalFacadeException {
    if (val.getId() <= LocationVO.maxReservedId) {
      // claim it doesn't exist
      return 1;
    }

    /** Only allow delete if not in use
     */
    if (getCal().checkLocationRefs(val)) {
      return 2;
    }

    /* Remove from preferences */
    getCal().getUserAuth().removeLocation(null, val);

    /* Attempt to delete
     */
/*    if (pars.getPublicAdmin()) {
      if (getCal().deletePublicLocation(val)) {
        getCal().touch();
        return 0;
      }
    } else */
    if (getCal().deleteLocation(val)) {
      return 0;
    }

    return 1;
  }

  /** Ensure a location exists. If it does sets the id in the object and returns
   * false (not added).
   * If not creates the entity then sets the new id in the object and returns
   * true (added).
   *
   * @param location   LocationVO value object. If this object has the id set,
   *                  we assume the check was made previously.
   * @return boolean  true if we added the entry
   */
  public boolean ensureLocationExists(LocationVO location)
      throws CalFacadeException {
    if (location.getId() > 0) {
      // We've already checked it exists
      return false;
    }

    // Assume a new location
    LocationVO l = findLocation(location);

    if (l != null) {
      location.setId(l.getId());
      return false;
    }

    // doesn't exist at this point, so we add it to db table
    location.setIsPublic(pars.getPublicAdmin());

    getCal().addLocation(location);

    return true;
  }

  /** ===================================================================
   *                   Public Locations
   *  =================================================================== */

  /** Return public locations - excluding reserved entries. Returns an empty
   * collection for no locations.
   *
   * @return Collection        of location value objects
   */
  public Collection getPublicLocations() throws CalFacadeException {
    return getCal().getPublicLocations();
  }

  /** Return given public location
   *
   * @param id          int id of the location
   * @return LocationVo object representing the location in question
   *                     null if it doesn't exist.
   */
  public LocationVO getPublicLocation(int id) throws CalFacadeException {
    return getCal().getPublicLocation(id);
  }

  /** ===================================================================
   *                   Sponsors
   *  =================================================================== */

  /** Return a collection of SponsorVO objects. Returns an empty
   * collection for no sponsors.
   *
   * @return Collection     of sponsors
   */
  public Collection getSponsors() throws CalFacadeException {
    return getCal().getSponsors();
  }

  /** Add a sponsor to the database. The id will be set in the parameter
   * object.
   *
   * @param val   SponsorVO object to be added
   */
  public void addSponsor(SponsorVO val) throws CalFacadeException {
    getCal().addSponsor(val);
  }

  /** Replace a sponsor in the database.
   *
   * @param val   SponsorVO object to be replaced
   */
  public void replaceSponsor(SponsorVO val) throws CalFacadeException {
    getCal().replaceSponsor(val);
  }

  /** Return sponsor with the given id
   *
   * @param id         int id of the sponsor
   * @return SponsorVo object representing the sponsor in question
   *                     null if it doesn't exist.
   */
  public SponsorVO getSponsor(int id) throws CalFacadeException {
    return getCal().getSponsor(id);
  }

  /** Search the public sponsors for a matching name, phone, email and link
   *
   * @param s          SponsorVO object to match for
   * @return SponsorVO matching object or null if no match.
   */
  public SponsorVO findSponsor(SponsorVO s) throws CalFacadeException {
    Collection ss = getSponsors();

    if (ss.isEmpty()) {
      return null;
    }

    String name = s.getName();
    String phone = s.getPhone();
    String email = s.getEmail();
    String link = s.getLink();

    Iterator it = ss.iterator();

    while (it.hasNext()) {
      SponsorVO si = (SponsorVO)it.next();

      if ((name.equals(si.getName())) &&
          checkField(phone, si.getPhone()) &&
          checkField(email, si.getEmail()) &&
          checkField(link, si.getLink())) {
        return si;
      }
    }

    return null;
  }

  /** Delete a sponsor with the given id. Also remove it from the current
   * user preferences (if any).
   *
   * @param val      SponsorVO object to be deleted
   * @return int     0 if it was deleted.
   *                 1 if it didn't exist
   *                 2 if in use
   */
  public int deleteSponsor(SponsorVO val) throws CalFacadeException {
    /** Only allow delete if not in use
     */
    if (getCal().checkSponsorRefs(val)) {
      return 2;
    }

    /* Remove from preferences */
    getCal().getUserAuth().removeSponsor(null, val);

    /* Attempt to delete
     */
    if (getCal().deleteSponsor(val)) {
      return 0;
    }

    return 1; //doesn't exist
  }

  /** Ensure a sponsor exists. If it does sets the id in the object and returns
   * false (not added).
   * If not creates the entity then sets the new id in the object and returns
   * true (added).
   *
   * @param sponsor   SponsorVO value object. If this object has the id set,
   *                  we assume the check was made previously.
   * @return boolean  true if we added the entry
   */
  public boolean ensureSponsorExists(SponsorVO sponsor)
      throws CalFacadeException {
    if (sponsor.getId() > 0) {
      // We've already checked it exists
      return false;
    }

    // Assume user entered a new sponsor
    SponsorVO s = findSponsor(sponsor);

    if (s != null) {
      sponsor.setId(s.getId());
      return false;
    }

    // doesn't exist at this point, so we add it to db table
    sponsor.setIsPublic(true);

    getCal().addSponsor(sponsor);

    return true;
  }

  /** ===================================================================
   *                   Public Sponsors
   *  =================================================================== */

  /** Return public sponsors - excluding reserved entries. Returns an empty
   * collection for no sponsors.
   *
   * @return Collection        of sponsor value objects
   */
  public Collection getPublicSponsors() throws CalFacadeException {
    return getCal().getPublicSponsors();
  }

  /** Return given public sponsor
   *
   * @param id         int id of the sponsor
   * @return SponsorVo object representing the sponsor in question
   *                     null if it doesn't exist.
   */
  public SponsorVO getPublicSponsor(int id) throws CalFacadeException {
    return getCal().getPublicSponsor(id);
  }

  /* ====================================================================
   *                   Events
   * ==================================================================== */

  /** Return a single event for the current user
   *
   * @param   eventId   int id of the event
   * @return  EventVO   value object representing event.
   */
  public EventVO getEvent(int eventId) throws CalFacadeException {
    return getCal().getEvent(eventId);
  }

  /** Return filtered events for the current user.
   *
   * @return Collection of event value objects
   */
  public Collection getEvents() throws CalFacadeException {
    return getCal().getEvents();
  }

  /** Delete an event. Optionally delete the event location if it is no
   * longer in use.
   *
   * @param event              EventVO object to be deleted. Null results in
   *                           a nothing deleted return.
   * @param delUnreffedLoc     delete the location if unreferenced
   * @return DelEventResult    result.
   * @exception CalFacadeException If there's a database access problem
   */
  public DelEventResult deleteEvent(EventVO event,
                                    boolean delUnreffedLoc) throws CalFacadeException {
    if (event == null) {
      return new DelEventResult(false, false);
    }

    boolean deletedLoc = false;
    LocationVO loc = event.getLocation();

    if (!getCal().deleteEvent(event)) {
      return new DelEventResult(false, false);
    }

    if (delUnreffedLoc) {
      if ((loc != null) &&
          (!getCal().checkLocationRefs(loc)) &&
          (getCal().deleteLocation(loc))) {
        deletedLoc = true;
      }
    }

    return new DelEventResult(true, deletedLoc);
  }

  /** Return the events for the given day as an array of value objects
   *
   * @param   date        MyCalendarVO object defining day
   * @return  Collection  one days events or null for no events.
   */
  public Collection getDaysEvents(MyCalendarVO date) throws CalFacadeException {
    return getCal().getDaysEvents(date.getTime());
  }

  /** Add an event and ensure its location and sponsor exist.
   *
   * <p>For public events some calendar implementors choose to allow the
   * dynamic creation of locations and sponsors. For each of those, if we have
   * an id, then the object represents a preexisting database item.
   *
   * <p>Otherwise the client has provided information which will be used to
   * locate an already existing location or sponsor. Failing that we use the
   * information to create a new entry.
   *
   * <p>For user clients, we generally assume no sponsor and the location is
   * optional. However, both conditions are enforced at the application level.
   *
   * <p>On return the event objct will have been updated. In addition the
   * location and sponsor may have been updated.
   *
   * @param event        EventVO object to be added
   * @return EventResult Counts of changes.
   */
  public EventUpdateResult addEvent( EventVO event) throws CalFacadeException {
    EventUpdateResult updResult = new EventUpdateResult();
    LocationVO loc = event.getLocation();
    SponsorVO sp = event.getSponsor();

    if ((sp != null) && ensureSponsorExists(sp)) {
      updResult.sponsorsAdded++;
    }

    if ((loc != null) && ensureLocationExists(loc)) {
      updResult.locationsAdded++;
    }

    event.setPublic(pars.getPublicAdmin());
    event.setCreator(getCal().getUserVO());
    Timestamp now = new Timestamp(System.currentTimeMillis());

    event.setLastmod(now);
    event.setCreated(now);

    getCal().addEvent(event);

    getCal().fixKeywords(event, null);
    /*
    // finally, add the eventkeywords
    Iterator it = event.iterateKeywords();

    while (it.hasNext()) {
      KeywordVO k = (KeywordVO)it.next();

      getCal().addEventKeyword(event, k);
    }
    */

    return updResult;
  }

  /** Update an event and ensure referenced entities such as location and
   * contact exist. The same approach is taken with these as for addEvent wth the
   * following addition.
   *
   * <p>If the location or sponsor changes, the old ones may no longer be
   * referenced by any events. This method will delete the old unused
   * entries if the global parameters specify that behaviour.
   *
   * <p>The original event can be supplied. If not a copy will be retrieved
   * before the updates are applied.
   *
   * @param event         updated EventVO object
   * @param oldEvent      EventVO object before changes applied
   * @return EventUpdateResult Counts of changes.
   */
  public EventUpdateResult updateEvent(EventVO event,
                                       EventVO oldEvent)
          throws CalFacadeException {
    EventUpdateResult updResult = new EventUpdateResult();
    if (oldEvent == null) {
      oldEvent = getEvent(event.getId());
      if (oldEvent == null) {
        throw new CalFacadeException("Event does not exist: " + event.getId());
      }
    }

    SponsorVO sp = event.getSponsor();

    if ((sp != null) && ensureSponsorExists(sp)) {
      updResult.sponsorsAdded++;
    }

    LocationVO loc = event.getLocation();

    if ((loc != null) && ensureLocationExists(loc)) {
      updResult.locationsAdded++;
    }

    getCal().updateEvent(event);

    /* If keywords changed, we need to change event_keywords entry
     * /

    Iterator it = event.iterateKeywords();

    while (it.hasNext()) {
      KeywordVO k = (KeywordVO)it.next();

      if (!oldEvent.hasKeyword(k)) {
        // Added the category
        getCal().addEventKeyword(event, k);
      }
    }

    it = oldEvent.iterateKeywords();

    while (it.hasNext()) {
      KeywordVO k = (KeywordVO)it.next();

      if (!event.hasKeyword(k)) {
        // Deleted the category
        getCal().deleteEventKeyword(event, k);
      }
    }*/
    getCal().fixKeywords(event, oldEvent);

    /** If we updated sponsor and location we might want to delete the
        old ones. */

    LocationVO origLocation = oldEvent.getLocation();
    SponsorVO origSponsor = oldEvent.getSponsor();

    boolean deleteOldLoc = !pars.getPublicAdmin() ||
                           pars.getAutoDeleteLocations();

    /* If we auto delete and the location changed and if the old location
       id no longer exists in events table then delete that location.
       */
    if (deleteOldLoc && (origLocation != null) &&
        (origLocation.getId() != event.getLocation().getId()) &&
        !getCal().checkLocationRefs(origLocation)) {
      getCal().deleteLocation(origLocation);
      updResult.locationsRemoved++;
    }

    boolean deleteOldSp = !pars.getPublicAdmin() ||
                          pars.getAutoDeleteSponsors();

    /* If we auto delete and the sponsor changed and if the old sponsorid
       no longer exists in events table then delete that sponsor.
       */
    if (deleteOldSp && (origSponsor != null) &&
        (origSponsor.getId() != event.getSponsor().getId()) &&
        !getCal().checkSponsorRefs(origSponsor)) {
      getCal().deleteSponsor(origSponsor);
      updResult.sponsorsRemoved++;
    }

    return updResult;
  }

  /* ====================================================================
   *                   Public Events
   * ==================================================================== */

  /** Return public events for the current user which match the supplied
   * alert flag
   *
   * @param alerts      boolean true if we want alerts false for exclude alerts
   * @return Collection  populated event value objects
   */
  public Collection getPublicEvents(boolean alerts)
      throws CalFacadeException {
    return getPublicEvents(alerts, false);
  }

  /** Return active public events for the current user which match the
   * supplied alert flag
   *
   * @param alerts      boolean true if we want alerts false for exclude alerts
   * @return Collection  populated event value objects
   */
  public Collection getActivePublicEvents(boolean alerts)
      throws CalFacadeException {
    return getPublicEvents(alerts, true);
  }

  /* * Return the public events for the given day as an array of value objects
   *
   * @param   date    MyCalendar object defining day
   * @return  EventVO[]  one days events or null for no events.
   * /
  public EventVO[] getDaysPublicEvents(MyCalendarVO date)
      throws CalFacadeException {
    return getCal().getDaysPublicEvents(date);
  }*/

  /* * Return the alerts for the given day as an array of value objects
   *
   * @param   date    MyCalendarVO object defining day
   * @return  EventVO[]  one days alerts or null for no events.
   * /
  public EventVO[] getDaysAlerts(MyCalendarVO date) throws CalFacadeException {
    EventVO[] evs = getCal().getDaysPublicEvents(date);

    if (evs == null) {
      return new EventVO[0];
    }

    Vector v = new Vector();

    for (int i = 0; i < evs.length; i++) {
      if (getKeyAlert(evs[i].getKeyword(0).getId())) {
        v.addElement(evs[i]);
      }
    }

    return (EventVO[])v.toArray(new EventVO[v.size()]);
  }*/

  /* * Return the public events for the current user. This is an administrative
   * function for the public events admin client.
   *
   * @return  EventVO[]  events or null for no events.
   * /
  public EventVO[] findPublicEvents() throws CalFacadeException {
    //return getCal().findPublicEvents(creator);
    throw new CalFacadeException("No direct access to method");
  }*/

  /* * Return the public events for the current user within the given date
   * range. This is an administrative function for the public events admin
   * client.
   *
   * @return  EventVO[]  events or null for no events.
   * /
  public EventVO[] findPublicEvents(MyCalendarVO startDate,
                                    MyCalendarVO endDate)
          throws CalFacadeException {
    //return getCal().findPublicEvents(creator, startDate, endDate);
    throw new CalFacadeException("No direct access to method");
  }*/

  /** Delete an event and optionally delete the old location and/or sponsor
   * entries if now unused.
   *
   * @param deleteOldLoc  true if we should delete old unused location
   * @param deleteOldSp  true if we should delete old unused sponsor
   * @param event      EventVO object to be added
   */
  public void deletePublicEvent(boolean deleteOldLoc,
                                boolean deleteOldSp,
                                EventVO event) throws CalFacadeException {
    getCal().deleteEvent(event);

    /* If we auto delete and if sponsor id no longer exists in events
       table then delete this sponsor.
     */
    if (deleteOldLoc &&
        !getCal().checkSponsorRefs(event.getSponsor())) {
      getCal().deleteSponsor(event.getSponsor());
    }

    /* If we auto delete and if location id no longer exists in events
       table then delete this location.
     */
    if (deleteOldSp &&
        !getCal().checkLocationRefs(event.getLocation())) {
      getCal().deleteLocation(event.getLocation());
    }
  }

  /** Set a public location object in an event. Does not update anything.
   *
   * @param  locId          Keys of the desired location
   * @param  event          EventVO object
   * @return LocationVO     Location object or null if none with that key
   */
  public LocationVO setEventPublicLocation(int locId, EventVO event)
          throws CalFacadeException {
    LocationVO l = getCal().getPublicLocation(locId);

    if (l == null) {
      // Somebody's faking
      return null;
    }

    event.setLocation(l);

    return l;
  }

  /* ====================================================================
   *                   EventRefs
   * ==================================================================== */

  /** Create an event ref for the current user. The event must not be owned
   * by the current user.
   *
   * @param  event      EventVO object
   * @return  boolean   false for no such item.
   */
  public boolean addEventRef(EventVO event) throws CalFacadeException {
    if (!event.isPublic()) {
      // When we get groups we need to widen the possibilities
      throw new CalFacadeException(
            "Attempt to create a reference to non-public event: " +
            event);
    }
    return getCal().addEventRef(event);
  }

  /** Delete an event ref for the current user. The event must be the target
   * of a reference.
   *
   * @param  event          EventVO object
   * @return  boolean       false for no such item.
   */
  public boolean deleteEventRef(EventVO event) throws CalFacadeException {
    if (event.getKind() != EventVO.kindAdded) {
      throw new CalFacadeException(
            "Attempt to delete a reference to non-referenced event: " +
            event);
    }
    return getCal().deleteEventRef(event);
  }

  /** Delete a subscribed event for the current user.
   *
   * <p>This is a partially implemented feature which currently only changes
   * the synch state. What we need is in effect a reverse event ref - I guess
   * a new purpose, which indicates the event is hidden for this user.
   *
   * @param  event          EventVO object
   * @return  boolean       false for no such item.
   */
  public boolean deleteSubscribedEvent(EventVO event) throws CalFacadeException {
    if (event.getKind() != EventVO.kindUndeletable) {
      throw new CalFacadeException(
            "Attempt to hide a non-subscribed event: " + event);
    }
    return getCal().deleteSubscribedEvent(event);
  }


  /* ====================================================================
   *                   Synchronization
   * ==================================================================== */

  /** Get the synch info for the current device and user.
   *
   * @return SynchInfoVO object or null.
   */
   public SynchInfoVO getSynchInfo() throws CalFacadeException {
    return getCal().getSynchInfo();
   }

  /** Add the synch info for the current device and user.
   *
   * @param val    SynchInfoVO object to add
   */
   public void addSynchInfo(SynchInfoVO val) throws CalFacadeException {
    getCal().addSynchInfo(val);
   }

  /** Update the synch info for the current device and user. There must be
   * exactly one object to update.
   *
   * @param val    SynchInfoVO object to update
   */
   public void updateSynchInfo(SynchInfoVO val) throws CalFacadeException {
    getCal().updateSynchInfo(val);
   }

  /** Return synchronization state for the given event in the current
   * synchronization context as defined by the initialisation parameters,
   * that is the user and synchId. If the event has never been involved in a
   * synch process there will be no related synch state object.
   *
   * @param ev            EventVO object .
   * @return SynchStateVO object or null.
   */
  public SynchStateVO getSynchState(EventVO ev)
      throws CalFacadeException {
    return getCal().getSynchState(ev);
  }

  /** Get the deleted synch states for the current user
   *
   * @returns Collection of synch state objects (possibly empty)
   */
  public Collection getDeletedSynchStates() throws CalFacadeException {
    return getCal().getDeletedSynchStates();
  }

  /** Add the synch state for the given event and user.
   *
   * @param val      SynchStateVO object.
   */
  public void addSynchState(SynchStateVO val)
      throws CalFacadeException {
    getCal().addSynchState(val);
  }

  /** Update the synch state for the given event and user. There must be
   * exactly one object to update.
   *
   * @param val    SynchStateVO object with associated event and user
   */
  public void updateSynchState(SynchStateVO val)
      throws CalFacadeException {
    getCal().updateSynchState(val);
  }

  /** Get the synch data associated with the given object..
   *
   * @param val    SynchStateVO object
   */
  public void getSynchData(SynchStateVO val) throws CalFacadeException {
    getCal().getSynchData(val);
  }

  /** Set the synch state and data for the given user and event.
   * There may be zero to many objects to update.
   *
   * @param val    SynchStateVO object with associated event and user
   */
  public void setSynchData(SynchStateVO val) throws CalFacadeException {
    getCal().setSynchData(val);
  }

  /** Update the synch states for the current user and device.
   *
   * <p>This is called at the termination of synchronization to set any state
   * objects appropriately.
   *
   * <p><table cols="2">
   * <th><td>Old state</td><td>New state</td></th>
   * <tr><td>UNKNOWN</td><td>SYNCHRONIZED</td></tr>
   * <tr><td>SYNCHRONIZED</td><td>SYNCHRONIZED</td></tr>
   * <tr><td>NEW</td><td>SYNCHRONIZED</td></tr>
   * <tr><td>MODIFIED</td><td>SYNCHRONIZED</td></tr>
   * <tr><td>DELETED</td><td>entry deleted</td></tr>
   * <tr><td>CLIENT_DELETED</td><td>unchanged</td></tr>
   * <tr><td>CLIENT_DELETED_UNDELIVERED</td><td>CLIENT_DELETED</td></tr>
   * <tr><td>CLIENT_MODIFIED</td><td>unchanged</td></tr>
   * </table>
   */
  public void updateSynchStates() throws CalFacadeException {
    getCal().updateSynchStates();
  }

  /* ====================================================================
   *                       Recurrences
   * This is currently under development at RPI
   * ==================================================================== */

  /** Add a recurrence to the database.
   *
   * @param val      RecurVO object.
   */
  public void addRecur(RecurVO val) throws CalFacadeException {
    getCal().addRecur(val);
  }

  /** Get a recurrence associated with an event.
   *
   * @param ev      Associated EventVO
   * @return RecurVO  recurrence object or null
   */
  public RecurVO getRecur(EventVO ev) throws CalFacadeException {
    return getCal().getRecur(ev);
  }

  /** Remove a recurrence associated with an event.
   *
   * @param ev      Associated EventVO
   */
  public void deleteRecur(EventVO ev) throws CalFacadeException {
    getCal().deleteRecur(ev);
  }

  /** Update a recurrence to the database.
   *
   * @param val      RecurVO object.
   */
  public void updateRecur(RecurVO val) throws CalFacadeException {
    getCal().updateRecur(val);
  }

  /* ====================================================================
   *                   Private methods
   * ==================================================================== */

  /** Add a keyword attribute name/value pair for a keyword
   *
   * @param KeywordVO       keyword object
   * @param name            String attribute name
   * @param val             String attribute value
   */
  private void addKeywordAttr(KeywordVO k,
                              String name,
                              String val) throws CalFacadeException {
    if (k.addKeyAttr(name, val)) {
      getCal().addKeywordAttr(k, name, val);
    }
  }

  /** Remove a keyword attribute name/value pair for a keyword
   *
   * @param KeywordVO       keyword object
   * @param name            String attribute name
   * @param val             String attribute value
   */
  private void removeKeywordAttr(KeywordVO k,
                                 String name,
                                 String val) throws CalFacadeException {
    if (k.removeKeyAttr(name, val)) {
      getCal().removeKeywordAttr(k, name, val);
    }
  }

  private Collection getPublicEvents(boolean alerts,
                                    boolean active)
      throws CalFacadeException {
    Collection evs;

    if (active) {
      evs = getCal().findPublicEvents(new java.sql.Date(System.currentTimeMillis()), null);
    } else {
      evs = getCal().findPublicEvents();
    }

    Vector v = new Vector();

    Iterator it = evs.iterator();

    while (it.hasNext()) {
      EventVO ev = (EventVO)it.next();

      KeywordVO k = ev.getFirstKeyword();

      if ((k != null) &&
          (alerts == getKeyAlert(k))) {
        v.addElement(ev);
      }
    }

    return v;
  }

  /** Search the locations for a matching name. This is used to find a
   * location by name when creating or updating events.
   *
   * <p>For a non-admin user we only look in the users set of locations.
   *
   * <p>For an admin user we look in all public locations - not just those
   * created by this user.
   *
   * @param l          LocationVO object to match for
   * @return LocationVO matching object or null if no match.
   */
  private LocationVO findLocation(LocationVO l) throws CalFacadeException {
    Collection ll;

    if (pars.getPublicAdmin()) {
      ll = getPublicLocations();
    } else {
      ll = getLocations();
    }

    if (ll.isEmpty()) {
      return null;
    }

    String address = l.getAddress();
    String subaddress = l.getSubaddress();
    String link = l.getLink();

    Iterator it = ll.iterator();

    while (it.hasNext()) {
      LocationVO li = (LocationVO)it.next();

      if ((address.equals(li.getAddress())) &&
          checkField(subaddress, li.getSubaddress()) &&
          checkField(link, li.getLink())) {
        return li;
      }
    }

    return null;
  }

  private Calintf getCal() throws CalFacadeException {
    if (cali == null) {
      cali = CalintfFactory.getCalintf(pars.getAuthUser(), pars.getRights(),
                                       pars.getUser(),
                                       pars.getAuthUser() == null,
                                       pars.getPublicAdmin(),
                                       pars.getSynchId(), debug);
    }

    return cali;
  }

  private boolean checkField(String fld1, String fld2) {
    if (fld1 == null) {
      if (fld2 == null) {
        return true;
      }
    } else if (fld1.equals(fld2)) {
      return true;
    }

    return false;
  }

  /** Get a logger for messages
   */
  private Logger getLogger() {
    if (log == null) {
      log = Logger.getLogger(this.getClass());
    }

    return log;
  }

  private void trace(String msg) {
    getLogger().debug("trace: " + msg);
  }
}

