/* **********************************************************************
    Copyright 2003 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.rpi.cct.uwcal.calsvci;

import edu.rpi.cct.uwcal.resources.Resources;

import edu.washington.cac.calfacade.shared.AdminGroups;
import edu.washington.cac.calfacade.shared.CalFacadeException;
import edu.washington.cac.calfacade.shared.CalendarVO;
import edu.washington.cac.calfacade.shared.EventVO;
import edu.washington.cac.calfacade.shared.filter.FilterVO;
import edu.washington.cac.calfacade.shared.KeywordVO;
import edu.washington.cac.calfacade.shared.LocationVO;
import edu.washington.cac.calfacade.shared.MyCalendarVO;
import edu.washington.cac.calfacade.shared.RecurVO;
import edu.washington.cac.calfacade.shared.SponsorVO;
import edu.washington.cac.calfacade.shared.SynchInfoVO;
import edu.washington.cac.calfacade.shared.SynchStateVO;
import edu.washington.cac.calfacade.shared.UserAuth;
import edu.washington.cac.calfacade.shared.UserVO;

import java.io.Serializable;
import java.sql.Timestamp;
import java.util.Collection;
import java.util.TreeSet;

/** This is the service interface to the calendar suite. This will be
 * used by web applications and web services as well as other applications
 * which wish to integrate calendar actions into their interface.
 *
 * <p>This is a high level interface which carries out commonly used
 * calendar operations which may involve a number of interactions with the
 * underlying database implementation.
 *
 * @author Mike Douglass       douglm@rpi.edu
 */
public interface CalSvcI extends Serializable {
  /** (Re)initialise the object for a particular use.
   *
   * @param pars        Defines the global parameters for the object
   * @return Calintf    object for interactions with back-end.
   */
  public void init(CalSvcIPars pars) throws CalFacadeException;

  /** Switch to the given non-null user. The access rights are determined by
   * the previous call to init.
   *
   * @param val         String user id
   */
  public void setUser(String val) throws CalFacadeException;

  /** Signal the start of a sequence of operations. These overlap transactions
   * in that there may be 0 to many transactions started and ended within an
   * open/close call and many open/close calls within a transaction.
   *
   * <p>The open close methods are mainly associated with web style requests
   * and open will usually be called early in the incoming request handling
   * and close will always be called on the way out allowing the interface an
   * opportunity to reacquire (on open) and release (on close) any resources
   * such as connections.
   */
  public void open() throws CalFacadeException;

  /** Call on the way out after handling a request..
   */
  public void close() throws CalFacadeException;

  /** Start a (possibly long-running) transaction. In the web environment
   * this might do nothing. The endTransaction method should in some way
   * check version numbers to detect concurrent updates and fail with an
   * exception.
   */
  public void beginTransaction() throws CalFacadeException;

  /** End a (possibly long-running) transaction. In the web environment
   * this should in some way check version numbers to detect concurrent updates
   * and fail with an exception.
   */
  public void endTransaction() throws CalFacadeException;

  /** Call if there has been an error during an update process.
   */
  public void rollbackTransaction() throws CalFacadeException;

  /** Get the set of internationalized resources for this session
   */
  public Resources getResources();

  /** Get a name uniquely.identifying this system. This should take the form <br/>
   *   name@host
   * <br/>where<ul>
   * <li>name identifies the particular calendar system at the site</li>
   * <li>host part identifies the domain of the site.</li>..
   * </ul>
   *
   * @return String    globally unique system identifier.
   */
  public String getSysid() throws CalFacadeException;

  /** Returns a value object representing the current user.
   *
   * @return UserVO       representing the current user
   */
  public UserVO getUserVO() throws CalFacadeException;

  /** Get a UserAuth object which allows the application to determine what
   * special rights the user has.
   *
   * <p>Note that the returned object may require a one time initialisation
   * call to set it up correctly.
   * @see edu.washington.cac.calfacade.shared.UserAuth#initialise(String, int)
   * @see edu.washington.cac.calfacade.shared.UserAuth#initialise(String, Object)
   *
   * @param user         String account name
   * @param par          parameter object
   * @return UserAuth    implementation.
   */
  public UserAuth getUserAuth(String user, Object par) throws CalFacadeException;

  /** Get an initialised UserAuth object
   *
   * @return UserAuth    implementation.
   */
  public UserAuth getUserAuth() throws CalFacadeException;

  /** Get an AdminGroups object which allows the application to handle
   * administrative groups.
   *
   * @return AdminGroups    implementation.
   */
  public AdminGroups getAdminGroups() throws CalFacadeException;

  /** Refresh the users events cache
   */
  public void refreshEvents() throws CalFacadeException;

  /** Return true if we believe a refresh is required. After a return of
   * true this method willreturn false until somebody updates the db.
   *
   * @return boolean    time to refresh
   */
  public boolean refreshNeeded() throws CalFacadeException;

  /* ====================================================================
   *                   Calendars and search
   * ==================================================================== */

  /** Return a list of calendar we can present to the client as a list of
   * available calendars.
   *
   * @return Collection   of CalendarVO
   */
  public Collection getCalendars() throws CalFacadeException;

  /** Set a search filter using the suppplied search string
   *
   * @param val    String search parameters
   */
  public void setSearch(String val) throws CalFacadeException;

  /** Return the current search string
   *
   * @return  String     search parameters
   */
  public String getSearch() throws CalFacadeException;

  /** Set the calendar we are interested in. This is represented by the name
   * of a calendar.
   *
   * @param  val     int id of calendar - 0 for none
   * @return boolean true for OK, false for unknown calendar
   */
  public boolean setCalendar(int val) throws CalFacadeException;

  /** Get the name of the current calendar we are displaying
   *
   * @return String    name of calendar or null for all events
   */
  public String getCalendarTitle() throws CalFacadeException;

  /** This method is called to add a calendar. No consistency checks are
   * carried out at the moment.
   *
   * <p>This is a temp fix until we have some sort of admin interface to
   * create and update calendar definitions.
   *
   * @param val  String values for db
   */
  public void addCalendarDef(String val) throws CalFacadeException;

  /** Add a filter to the database. All references must be set.
   */
  public void addFilter(FilterVO val) throws CalFacadeException;

  /** Update a filter
   *
   * @param  val           FilterVO object to upate
   * @exception CalFacadeException If there's a problem
   */
  public void updateFilter(FilterVO val) throws CalFacadeException;

  /* * Set a filter which returns events starting after a given Timestamp.
   *
   * @param val   Timestamp object.
   * /
  public void setLastmodGECalendar(Timestamp val) throws CalFacadeException;

  /* * Set a filter which returns events starting after a given date of the
   * form yyyymmdd.
   *
   * @param date   String of form yyyymmdd.
   * /
  public void setLastmodGEDateCalendar(String date) throws CalFacadeException;

  /* * Set a filter which returns events starting after a given date and time
   * of the form yyyymmddhhmm.
   *
   * @param dateTime   String of form yyyymmddhhmm.
   * /
  public void setLastmodGEDateTimeCalendar(String dateTime)
      throws CalFacadeException;
  */

  /* ====================================================================
   *                   Subscriptions
   * ==================================================================== */

  /** Ensure that the users subscriptions match a set of calendars
   *
   * @param crs      TreeSet of Integer calendar ids
   * @return int[]   Number of calendars subscribed to, and number of
   *                 calendars unsubscribed to
   */
  public int[] fixSubscriptions(TreeSet crs) throws CalFacadeException;

  /** Determine if user is subscribed to a calendar
   *
   * @param id       int calendar id
   * @return boolean true if subscribed to calendars
   */
  public boolean getSubscribed(int id) throws CalFacadeException;

  /* ====================================================================
   *                   Keywords
   * ==================================================================== */

  /** Add a Keyword to the database. The id will be set in the parameter
   * object.
   *
   * @param val   KeywordVO object to be added
   */
  public void addKeyword(KeywordVO val) throws CalFacadeException;

  /** Add a Keyword to the database and give it the assigned key.
   * Only for restores
   *
   * @param val   KeywordVO object to be added with key set
   */
  public void addKeywordWithKey(KeywordVO val) throws CalFacadeException;

  /** Replace a keyword in the database.
   *
   * @param val   KeywordVO object to be replaced
   */
  public void replaceKeyword(KeywordVO val) throws CalFacadeException;

  /** Delete a keyword with the given id. Also remove it from the current
   * user preferences (if any).
   *
   * @param val      KeywordVO keyword
   * @return int     0 if it was deleted.
   *                 1 if it didn't exist
   *                 2 if in use
   */
  public int deleteKeyword(KeywordVO val) throws CalFacadeException;

  /** Search the keywords owned by this user for a matching name
   *
   * @param k          KeywordVO object to match for
   * @return KeywordVO matching object or null if no match.
   */
  public KeywordVO findKeyword(KeywordVO k) throws CalFacadeException;

  /** Ensure a keyword exists. If it does sets the id in the object and returns
   * false (not added).
   * If not creates the entity then sets the new id in the object and returns
   * true (added)..
   *
   * @param keyword   KeywordVO value object. If this object has the id set,
   *                  we assume the check was made previously.
   * @return boolean  true if we added the entry
   */
  public boolean ensureKeywordExists(KeywordVO keyword) throws CalFacadeException;

  /* ====================================================================
   *                   Public Keywords
   * ==================================================================== */

  /** Return public keywords excluding alerts.
   *
   * @return Collection    of alert keywords
   */
  public Collection getPublicKeywords() throws CalFacadeException;

  /** Return alert keywords only.
   *
   * @return Collection    of alert keywords
   */
  public Collection getAlertKeywords() throws CalFacadeException;

  /** Return public keywords including alerts.
   *
   * @return Collection    of alert keywords
   */
  public Collection getAllPublicKeywords() throws CalFacadeException;

  /** Return given public keyword
   *
   * @param id          int id of the keyword
   * @return KeywordVO  keywords value object
   */
  public KeywordVO getPublicKeyword(int id) throws CalFacadeException;

  /* ====================================================================
   *                   Keyword attributes
   * ==================================================================== */

  /** Set keyword alert on/off
   *
   * @param KeywordVO  keyword object
   * @param val        boolean turn it on/off
   */
  public void setKeyAlert(KeywordVO k, boolean val) throws CalFacadeException;

  /** Is given keyword an alert?
   *
   * @param KeywordVO  keyword object
   * @return boolean   true if this is an alert keyword
   */
  public boolean getKeyAlert(KeywordVO k) throws CalFacadeException;

  /** ===================================================================
   *                   Locations
   *  =================================================================== */

  /** Return all locations for this user. Returns an empty collection for
   * no locations. If this is a public admin instance, this will return all
   * public locations owned by the current user. Otherwise it returns all
   * personal locations owned by the user.
   *
   * @return Collection        of location value objects
   */
  public Collection getLocations() throws CalFacadeException;

  /** Add a location to the database. The id will be set in the parameter
   * object.
   *
   * @param val   LocationVO object to be added
   */
  public void addLocation(LocationVO val) throws CalFacadeException;

  /** Replace a location in the database.
   *
   * @param val   LocationVO object to be replaced
   */
  public void replaceLocation(LocationVO val) throws CalFacadeException;

  /** Return a location with the given id
   *
   * @param id     int id of the location
   * @return LocationVo object representing the location in question
   *                     null if it doesn't exist.
   */
  public LocationVO getLocation(int id) throws CalFacadeException;

  /** Delete a given location. If this is public admin the location must be
   * public.
   *
   * @param val      LocationVO object to be deleted
   * @return int     0 if it was deleted.
   *                 1 if it didn't exist
   *                 2 if in use
   */
  public int deleteLocation(LocationVO val) throws CalFacadeException;

  /** Ensure a location exists. If it does sets the id in the object and returns
   * false (not added).
   * If not creates the entity then sets the new id in the object and returns
   * true (added).
   *
   * @param location   LocationVO value object. If this object has the id set,
   *                  we assume the check was made previously.
   * @return boolean  true if we added the entry
   */
  public boolean ensureLocationExists(LocationVO location)
      throws CalFacadeException;

  /* ====================================================================
   *                   Public Locations
   * ==================================================================== */

  /** Return public locations - excluding reserved entries. Returns an empty
   * collection for no locations.
   *
   * @return Collection        of location value objects
   */
  public Collection getPublicLocations() throws CalFacadeException;

  /** Return given public location
   *
   * @param id          int id of the location
   * @return LocationVo object representing the location in question
   *                     null if it doesn't exist.
   */
  public LocationVO getPublicLocation(int id) throws CalFacadeException;

  /* ====================================================================
   *                   Sponsors
   * ==================================================================== */

  /** Return all sponsors for this user. Returns an empty collection for
   * no sponsors. If this is a public admin instance, this will return all
   * public sponsors owned by the current user. Otherwise it returns all
   * personal sponsors owned by the user.
   *
   * <p>(Note the current implementation does not allow forthe creation of
   * personal sponsors.
   *
   * @return Collection    of sponsors
   */
  public Collection getSponsors() throws CalFacadeException;

  /** Add a sponsor to the database. The id will be set in the parameter
   * object.
   *
   * @param val   SponsorVO object to be added
   */
  public void addSponsor(SponsorVO val) throws CalFacadeException;

  /** Replace a sponsor in the database.
   *
   * @param val   SponsorVO object to be replaced
   */
  public void replaceSponsor(SponsorVO val) throws CalFacadeException;

  /** Return sponsor with the given id
   *
   * @param id         int id of the sponsor
   * @return SponsorVo object representing the sponsor in question
   *                     null if it doesn't exist.
   */
  public SponsorVO getSponsor(int id) throws CalFacadeException;

  /** Search the public sponsors for a matching name, phone, email and link
   *
   * @param s          SponsorVO object to match for
   * @return SponsorVO matching object or null if no match.
   */
  public SponsorVO findSponsor(SponsorVO s) throws CalFacadeException;

  /** Delete a sponsor with the given id. Also remove it from the current
   * user preferences (if any).
   *
   * @param val      SponsorVO object to be deleted
   * @return int     0 if it was deleted.
   *                 1 if it didn't exist
   *                 2 if in use
   */
  public int deleteSponsor(SponsorVO val) throws CalFacadeException;

  /** Ensure a sponsor exists. If it does sets the id in the object and returns
   * false (not added).
   * If not creates the entity then sets the new id in the object and returns
   * true (added).
   *
   * @param sponsor   SponsorVO value object. If this object has the id set,
   *                  we assume the check was made previously.
   * @return boolean  true if we added the entry
   */
  public boolean ensureSponsorExists(SponsorVO sponsor) throws CalFacadeException;

  /* ====================================================================
   *                   Public Sponsors
   * ==================================================================== */

  /** Return public sponsors - excluding reserved entries. Returns an empty
   * collection for no sponsors.
   *
   * @return Collection        of sponsor value objects
   */
  public Collection getPublicSponsors() throws CalFacadeException;

  /** Return given public sponsor
   *
   * @param id         int id of the sponsor
   * @return SponsorVo object representing the sponsor in question
   *                     null if it doesn't exist.
   */
  public SponsorVO getPublicSponsor(int id) throws CalFacadeException;

  /* ====================================================================
   *                   Events
   * ==================================================================== */

  /** Return a single event for the current user
   *
   * @param   eventId   int id of the event
   * @return  EventVO   value object representing event.
   */
  public EventVO getEvent(int eventId) throws CalFacadeException;

  public static class DelEventResult {
    public boolean eventDeleted;
    public boolean locationDeleted;

    public DelEventResult(boolean eventDeleted,
                          boolean locationDeleted) {
      this.eventDeleted = eventDeleted;
      this.locationDeleted = locationDeleted;
    }
  }

  /** Return filtered events for the current user.
   *
   * @return Collection of event value objects
   */
  public Collection getEvents() throws CalFacadeException;

  /** Delete an event. Optionally delete the event location if it is no
   * longer in use.
   *
   * @param event              EventVO object to be deleted
   * @param delUnreffedLoc     delete the location if unreferenced
   * @return DelEventResult    result.
   * @exception CalFacadeException If there's a database access problem
   */
  public DelEventResult deleteEvent(EventVO event,
                                    boolean delUnreffedLoc) throws CalFacadeException;

  /** Return the events for the given day as an array of value objects
   *
   * @param   date    MyCalendarVO object defining day
   * @return  Collection  one days events or null for no events.
   */
  public Collection getDaysEvents(MyCalendarVO date) throws CalFacadeException;

  /** This class allows add and update event to signal changes backk to the
   * caller.
   */
  public static class EventUpdateResult {
    public int locationsAdded;
    public int sponsorsAdded;
    public int keywordsAdded;

    public int locationsRemoved;
    public int sponsorsRemoved;
    public int keywordsRemoved;
  }

  /** Add an event and ensure its location and sponsor exist.
   *
   * <p>For public events some calendar implementors choose to allow the
   * dynamic creation of locations and sponsors. For each of those, if we have
   * an id, then the object represents a preexisting database item.
   *
   * <p>Otherwise the client has provided information which will be used to
   * locate an already existing location or sponsor. Failing that we use the
   * information to create a new entry.
   *
   * <p>For user clients, we generally assume no sponsor and the location is
   * optional. However, both conditions are enforced at the application level.
   *
   * <p>On return the event objct will have been updated. In addition the
   * location and sponsor may have been updated.
   *
   * @param event        EventVO object to be added
   * @return EventUpdateResult Counts of changes.
   */
  public EventUpdateResult addEvent(EventVO event) throws CalFacadeException;

  /** Update an event and ensure referenced entities such as location and
   * contact exist. The same approach is taken with these as for addEvent wth the
   * following addition.
   *
   * <p>If the location or sponsor changes, the old ones may no longer be
   * referenced by any events. This method will delete the old unused
   * entries if the global parameters specify that behaviour.
   *
   * <p>The original event can be supplied. If not a copy will be retrieved
   * before the updates are applied.
   *
   * <p>Any synch states will be set to modified. In additon, if this is a
   * synchronization instance, the users synch state will be set to
   * synchronized.
   *
   * @param event         updated EventVO object
   * @param oldEvent      EventVO object before changes applied
   * @return EventUpdateResult Counts of changes.
   */
  public EventUpdateResult updateEvent(EventVO event,
                                       EventVO oldEvent)
          throws CalFacadeException;

  /* ====================================================================
   *                   Public Events
   * ==================================================================== */

  /** Return public events for the current user which match the supplied
   * alert flag
   *
   * @param alerts       boolean true if we want alerts false for exclude alerts
   * @return Collection  populated event value objects
   */
  public Collection getPublicEvents(boolean alerts)
      throws CalFacadeException;

  /** Return active public events for the current user which match the
   * supplied alert flag
   *
   * @param alerts       boolean true if we want alerts false for exclude alerts
   * @return Collection  populated event value objects
   */
  public Collection getActivePublicEvents(boolean alerts)
      throws CalFacadeException;

  /* * Return the alerts for the given day as an array of value objects
   *
   * @param   date    MyCalendarVO object defining day
   * @return  EventVO[]  one days alerts or null for no events.
   * /
  public EventVO[] getDaysAlerts(MyCalendarVO date) throws CalFacadeException;*/

  /** Delete an event and optionally delete the old location and/or sponsor
   * entries if now unused.
   *
   * @param deleteOldLoc  true if we should delete old unused location
   * @param deleteOldSp  true if we should delete old unused sponsor
   * @param event      EventVO object to be added
   */
  public void deletePublicEvent(boolean deleteOldLoc,
                                boolean deleteOldSp,
                                EventVO event) throws CalFacadeException;

  /** Set a public location object in an event. Does not update anything.
   *
   * @param  locId          Keys of the desired location
   * @param  event          EventVO object
   * @return LocationVO     Location object or null if none with that key
   */
  public LocationVO setEventPublicLocation(int locId, EventVO event)
      throws CalFacadeException;

  /* ====================================================================
   *                   EventRefs
   * ==================================================================== */

  /** Create an event ref for the current user.
   *
   * @param  event      EventVO object
   * @return  boolean   false for no such item.
   */
  public boolean addEventRef(EventVO event) throws CalFacadeException;

  /** Delete an event ref for the current user.
   *
   * @param  event          EventVO object
   * @return  boolean       false for no such item.
   */
  public boolean deleteEventRef(EventVO event) throws CalFacadeException;

  /** Delete a subscribed event for the current user.
   *
   * <p>This is a partially implemented feature which currently only changes
   * the synch state. What we need is in effect a reverse event ref - I guess
   * a new purpose, which indicates the event is hidden for this user.
   *
   * @param  event          EventVO object
   * @return  boolean       false for no such item.
   */
  public boolean deleteSubscribedEvent(EventVO event) throws CalFacadeException;

  /* ====================================================================
   *                   Synchronization
   * ==================================================================== */

  /** Get the synch info for the current device and user.
   *
   * @return SynchInfoVO object or null.
   */
   public SynchInfoVO getSynchInfo() throws CalFacadeException;

  /** Add the synch info for the current device and user.
   *
   * @param val    SynchInfoVO object to add
   */
  public void addSynchInfo(SynchInfoVO val) throws CalFacadeException;

  /** Update the synch info for the current device and user. There must be
   * exactly one object to update.
   *
   * @param val    SynchInfoVO object to update
   */
  public void updateSynchInfo(SynchInfoVO val)
      throws CalFacadeException;

  /** Return synchronization state for the given event in the current
   * synchronization context as defined by the initialisation parameters,
   * that is the user and synchId. If the event has never been involved in a
   * synch process there will be no related synch state object.
   *
   * @param ev            EventVO object .
   * @return SynchStateVO object or null.
   */
  public SynchStateVO getSynchState(EventVO ev)
      throws CalFacadeException;

  /** Get the deleted synch states for the current user
   *
   * @returns Collection of synch state objects (possibly empty)
   */
  public Collection getDeletedSynchStates() throws CalFacadeException;

  /** Add the synch state for the given event and user.
   *
   * @param val      SynchStateVO object.
   */
  public void addSynchState(SynchStateVO val)
      throws CalFacadeException;

  /** Update the synch state for the given event and user. There must be
   * exactly one object to update.
   *
   * @param val    SynchStateVO object with associated event and user
   */
  public void updateSynchState(SynchStateVO val)
      throws CalFacadeException;

  /** Get the synch data associated with the given object..
   *
   * @param val    SynchStateVO object
   */
  public void getSynchData(SynchStateVO val) throws CalFacadeException;

  /** Set the synch state and data for the given user and event.
   * There may be zero to many objects to update.
   *
   * @param val    SynchStateVO object with associated event and user
   */
  public void setSynchData(SynchStateVO val) throws CalFacadeException;

  /** Update the synch states for the current user and device.
   *
   * <p>This is called at the termination of synchronization to set any state
   * objects appropriately.
   *
   * <p><table cols="2">
   * <th><td>Old state</td><td>New state</td></th>
   * <tr><td>UNKNOWN</td><td>SYNCHRONIZED</td></tr>
   * <tr><td>SYNCHRONIZED</td><td>SYNCHRONIZED</td></tr>
   * <tr><td>NEW</td><td>SYNCHRONIZED</td></tr>
   * <tr><td>MODIFIED</td><td>SYNCHRONIZED</td></tr>
   * <tr><td>DELETED</td><td>entry deleted</td></tr>
   * <tr><td>CLIENT_DELETED</td><td>unchanged</td></tr>
   * <tr><td>CLIENT_DELETED_UNDELIVERED</td><td>CLIENT_DELETED</td></tr>
   * <tr><td>CLIENT_MODIFIED</td><td>unchanged</td></tr>
   * <tr><td>CLIENT_MODIFIED_UNDELIVERED</td><td>CLIENT_MODIFIED</td></tr>
   * </table>
   */
  public void updateSynchStates() throws CalFacadeException;

  /* ====================================================================
   *                       Recurrences
   * This is currently under development at RPI
   * ==================================================================== */

  /** Add a recurrence to the database.
   *
   * @param val      RecurVO object.
   */
  public void addRecur(RecurVO val) throws CalFacadeException;

  /** Get a recurrence associated with an event.
   *
   * @param ev      Associated EventVO
   * @return RecurVO  recurrence object or null
   */
  public RecurVO getRecur(EventVO ev) throws CalFacadeException;

  /** Remove a recurrence associated with an event.
   *
   * @param ev      Associated EventVO
   */
  public void deleteRecur(EventVO ev) throws CalFacadeException;

  /** Update a recurrence to the database.
   *
   * @param val      RecurVO object.
   */
  public void updateRecur(RecurVO val) throws CalFacadeException;
}

