/* **********************************************************************
    Copyright 2003 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.rpi.cct.uwcal.admin;

import edu.rpi.cct.uwcal.webcommon.UWCalSession;

import edu.washington.cac.calfacade.shared.SponsorVO;

import javax.servlet.http.HttpServletRequest;

/** This action adds or modifies sponsor.
 *
 * Actions:
 * <p><ul>
 * <li><em>initAddSponsor</em>Forwards to "addSponsor" after setting form
 * sponsor to null. A new object will be created from any following rsponse.</li>
 * <li><em>listUpdateSponsors</em>Forwards to "listUpdSponsors" to provide
 * list of sponsors for update</li>
 * <li><em>listSponsors</em>Forwards to "listSponsors" to provide
 * list of sponsors for display</li>
 * <li><em>deleteSponsor</em>Other request parameters will have created a
 * sponsor value object with the id set. Forwards to "sponsorInUse",
 * "noSuchSponsor" or "done"</li>
 *
 * @author Mike Douglass   douglm@rpi.edu
 */
public class PESponsorAction extends PEAbstractAction {
  /** This is the routine which does the work.
   *
   * @param request   Needed to locate session
   * @param sess      UWCalSession calendar session object
   * @param form      Action form
   * @return String   forward name
   */
  public String doAction(HttpServletRequest request,
                         UWCalSession sess,
                         PEActionForm form) throws Throwable {
    /** Check access
     */
    if (!form.getAuthorisedUser()) {
      return "noAccess";
    }

    /** =================================================================
     *                   Initial sponsor actions
     *  ================================================================= */

    String reqpar = request.getParameter("initAddSponsor");

    if (reqpar != null) {
      /** Set the objects to null so we get new ones.
       */
      form.initFields();

      return "addSponsor";
    }

    reqpar = request.getParameter("listUpdateSponsors");

    if (reqpar != null) {
      return "listUpdSponsors";
    }

    reqpar = request.getParameter("listSponsors");

    if (reqpar != null) {
      return "listSponsors";
    }

    /** =================================================================
     *                   Sponsor update actions
     *  ================================================================= */

    reqpar = request.getParameter("deleteSponsor");
    if (reqpar != null) {
      return deleteSponsor(form);
    }

    reqpar = request.getParameter("addSponsor");

    if (reqpar != null) {
      if (!form.validateNewSponsor()) {
        return "retryAdd";
      }

      updateSponsor(form, true);
      return "done";
    }

    reqpar = request.getParameter("action");

    if (String.valueOf(reqpar).equals("fetch")) {
      /** User requested a sponsor from the list. Retrieve it, embed it in
       * the form so we can display the page
       */
      int id = form.getSponsor().getId();

      SponsorVO sponsor = form.getCalSvcI().getPublicSponsor(id);

      if (debug) {
        if (sponsor == null) {
          logIt("No sponsor with id " + id);
        } else {
          logIt("Retrieved sponsor " + sponsor.getId());
        }
      }

      form.setSponsor(sponsor);
      if (sponsor == null) {
        err.emit("org.ucal.calendar.pubevents.error.nosuchsponsor",
                 id);
        return "notFound";
      }

      return "updateSponsor";
    }

    /** We are just updating from the current form values.
     */
    if (!form.validateNewSponsor()) {
      return "retryMod";
    }

    return updateSponsor(form, false);
  }

  private String deleteSponsor(PEActionForm form) throws Throwable {
    SponsorVO sp = form.getSponsor();
    int delResult = form.getCalSvcI().deleteSponsor(sp);

    if (delResult == 2) {
      msg.emit("org.ucal.calendar.pubevents.message.sponsor.referenced");
      return "sponsorInUse";
    }

    if (delResult == 1) {
       err.emit("org.ucal.calendar.pubevents.error.nosuchsponsor", sp);
      return "noSuchSponsor";
    }

    msg.emit("org.ucal.calendar.pubevents.message.sponsor.deleted");

    return "done";
  }

  private String updateSponsor(PEActionForm form,
                               boolean add) throws Throwable {
    /* if a sponsor with the same name, phone, e-mail, link, and creator exist
       in sponsors table then retrieve its sponsorid, otherwise add this
       sponsor to the database and then retrieve its sponsorid.
     */

    SponsorVO sp = form.getSponsor();

    if (!add && (sp.getId() <= SponsorVO.maxReservedId)) {
      // claim it doesn't exist
      err.emit("org.ucal.calendar.pubevents.error.nosuchsponsor",
               sp.getId());
      return "noSuchSponsor";
    }

    if (add) {
      form.getCalSvcI().ensureSponsorExists(sp);
    } else {
      form.getCalSvcI().replaceSponsor(sp);
    }

    form.updatePrefs(null, sp, null);

    if (add) {
      msg.emit("org.ucal.calendar.pubevents.message.sponsor.added");
    } else {
      msg.emit("org.ucal.calendar.pubevents.message.sponsor.updated");
    }

    return "done";
  }
}

