package edu.washington.cac.calfacade.impl;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;

import edu.washington.cac.calfacade.shared.AdminGroups;
import edu.washington.cac.calfacade.shared.AdminGroupROVO;
import edu.washington.cac.calfacade.shared.AdminGroupVO;
import edu.washington.cac.calfacade.shared.AuthUserVO;
import edu.washington.cac.calfacade.shared.CalFacadeException;
import edu.washington.cac.calfacade.shared.Calintf;
import edu.washington.cac.calfacade.shared.UserAuth;
import edu.washington.cac.calfacade.shared.UserVO;

/** An implementation of AdminGroups which stores the groups in the calendar
 * database.
 *
 * @author Mike Douglass douglm@rpi.edu
 * @version 1.0
 */
public class AdminGroupsDbImpl extends EventData implements AdminGroups {
  private Calintf cal;

  private AdminGroupROVO currentGroup;

  private static final String groupTable = "adminGroups";
  private static final String groupMemberTable = "adminGroupMembers";

  private static final String fldGroupName = "groupname";
  private static final String fldGroupDesc = "description";
  private static final String fldGroupGroupOwner = "groupOwner";
  private static final String fldGroupOwnerid = "ownerid";
  private static final String fldUserid = "userid";

  /** SQL to delete a group */
  private static final String deleteGroupSQL =
      "delete from " + groupTable + " where " +
        fldGroupName + "=?";

  /** SQL to add an entry to the group table */
  private static final String addGroupSQL =
      "insert into " + groupTable + " (" +
          fldGroupName + ", " +
          fldGroupDesc + ", " +
          fldGroupGroupOwner + ", " +
          fldGroupOwnerid + ") values (?, ?, ?, ?)";

  /** SQL to find an entry in the group table */
  private static final String findGroupSQL =
      "select " +
          fldGroupName + ", " +
          fldGroupDesc + ", " +
          fldGroupGroupOwner + ", " +
          fldGroupOwnerid + " from " +
        groupTable + " where " +
          fldGroupName + "=? ";

  /** SQL to find an entry in the group table with the given event owner. */
  private static final String findGroupByEventOwnerSQL =
      "select " +
          fldGroupName + ", " +
          fldGroupDesc + ", " +
          fldGroupGroupOwner + ", " +
          fldGroupOwnerid + " from " +
        groupTable + " where " +
          fldGroupOwnerid + "=? ";

  /** SQL to return all entries in the group table */
  private static final String allGroupsSQL =
      "select " +
          fldGroupName + ", " +
          fldGroupDesc + ", " +
          fldGroupGroupOwner + ", " +
          fldGroupOwnerid + " from " +
        groupTable + " order by " +
          fldGroupName;

  /** SQL to update an entry in the group table */
  private static final String updateGroupSQL =
      "update " + groupTable + " set " +
          fldGroupDesc + "=?, " +
          fldGroupGroupOwner + "=?, " +
          fldGroupOwnerid + "=? where " +
          fldGroupName + "=? ";

  /** SQL to delete a group member */
  private static final String deleteGroupMemberSQL =
      "delete from " + groupMemberTable + " where " +
        fldGroupName + "=? and " +
        fldUserid + "=?";

  /** SQL to delete all group members */
  private static final String deleteAllGroupMembersSQL =
      "delete from " + groupMemberTable + " where " +
        fldGroupName + "=?";

  /** SQL to add an entry to the group member table */
  private static final String addGroupMemberSQL =
      "insert into " + groupMemberTable + " (" +
          fldGroupName + ", " +
          fldUserid +") values (?, ?)";

  /** SQL to return members of a group */
  private static final String findGroupMembersSQL =
      "select " +
         fldUserid + " from " +
          groupMemberTable + " where " +
            fldGroupName + "=?";

  /** SQL to return names of groups for a user */
  private static final String findMemberGroupsSQL =
      "select " +
         fldGroupName + " from " +
          groupMemberTable + " where " +
            fldUserid + "=?";

  public void init(Calintf cal) {
    this.cal = cal;
  }

  /** ===================================================================
   *  The following return the state of the current user.
   *  =================================================================== */

  /** Set the current user group or null if the user is not (acting as) a
   * group member.
   *
   * @param name            String name of the group.
   * @exception CalFacadeException If there's a problem
   */
  public void setAdminGroup(String name) throws CalFacadeException {
    if (name == null) {
      currentGroup = null;
    } else {
      AdminGroupVO ug = findGroup(name);

      if (ug == null) {
        throw new CalFacadeException("Group " + name + " does not exist.");
      }

      currentGroup = ug.getAdminGroupROVO();
    }
  }

  /** Return a read only object representing the state of the current user
   * group. This will be null if the user is not in any group.
   *
   * @return AdminGroupVORO        object representing user state
   * @exception CalFacadeException If there's a problem
   */
  public AdminGroupROVO getAdminGroup() throws CalFacadeException {
    return currentGroup;
  }

  /** ===================================================================
   *  The following should not change the state of the current users
   *  group which is set and retrieved with the above methods.
   * The first group are available even if group maintenance is not.
   *  =================================================================== */

  /** Return all admin groups of which user is a member. Never returns null.
   *
   * @param user           UserVO object
   * @return Collection    of AdminGroupVO
   */
  public Collection getAdminGroups(String userid)
      throws CalFacadeException {
    PreparedStatement s = null;
    ResultSet rs = null;
    Vector v = new Vector();
    Vector ags = new Vector();

    try {
      if (userid != null) {
        s = getPreparedStatement(findMemberGroupsSQL);

        s.setString(1, userid);
      } else {
        s = getPreparedStatement(allGroupsSQL);
      }

      rs = s.executeQuery();

      while (rs.next()) {
        v.addElement(rs.getString(fldGroupName));
      }

      int sz = v.size();

      if (sz == 0) {
        return ags;
      }

      Iterator it = v.iterator();

      while (it.hasNext()) {
        ags.add(findGroup((String)it.next()));
      }

      /** Add the members
       */
      it = ags.iterator();

      while (it.hasNext()) {
        AdminGroupVO ag = (AdminGroupVO)it.next();

        ag.setGroupMembers(getMembers(ag.getName()));
      }

      return ags;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      if (rs != null) {
        try {
          rs.close();
        } catch (Throwable t) {}
        rs = null;
      }

      close(); // Will free s
      s = null;
    }
  }

  /** Find a group.
   *
   * @param  name          String group name
   * @return AdminGroupVO   group object
   * @exception CalFacadeException If there's a problem
   */
  public AdminGroupVO findGroup(String name) throws CalFacadeException {
    PreparedStatement s = null;
    ResultSet rs = null;

    try {
      s = getPreparedStatement(findGroupSQL);

      s.setString(1, name);
      rs = s.executeQuery();

      return makeGroupVO(rs);
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      if (rs != null) {
        try {
          rs.close();
        } catch (Throwable t) {}
        rs = null;
      }

      close(); // Will free s
      s = null;
    }
  }

  /** Find a group given the event owner.
   *
   * @param  id             String event owner id
   * @return AdminGroupVO   group object
   * @exception CalFacadeException If there's a problem
   */
  public AdminGroupVO findGroupByEventOwner(String id)
      throws CalFacadeException {
    PreparedStatement s = null;
    ResultSet rs = null;

    try {
      s = getPreparedStatement(findGroupByEventOwnerSQL);

      s.setString(1, id);
      rs = s.executeQuery();

      return makeGroupVO(rs);
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      if (rs != null) {
        try {
          rs.close();
        } catch (Throwable t) {}
        rs = null;
      }

      close(); // Will free s
      s = null;
    }
  }

  /** Show whether user entries can be modified with this
   * class. Some sites may use other mechanisms.
   *
   * @return boolean    true if group maintenance is implemented.
   */
  public boolean getGroupMaintOK() {
    return true;
  }

  /** ===================================================================
   *  The following are available if group maintenance is on.
   *  =================================================================== */

  /** Add an admin group
   *
   * @param  group           AdminGroupVO group object to add
   * @return AdminGroupVO   group object to be used instead of param
   * @exception CalFacadeException If there's a problem
   */
  public AdminGroupVO addGroup(AdminGroupVO group)
          throws CalFacadeException {
    PreparedStatement s = null;

    try {
      /* First ensure owner and event owner exist
       */
      ensureUserExists(group.getGroupOwner());
      ensureUserExists(group.getOwner());

      s = getPreparedStatement(addGroupSQL);

      s.setString(1, group.getName());
      s.setString(2, group.getDescription());
      s.setString(3, group.getGroupOwner().getAccount());
      s.setString(4, group.getOwner().getAccount());
      s.executeUpdate();

      return group;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Add an admin group and give it the assigned key.
   * Only for restores
   *
   * @param  group           AdminGroupVO group object to add with key set
   * @exception CalFacadeException If there's a problem
   */
  public void addGroupWithKey(AdminGroupVO group)
          throws CalFacadeException {
    throw new CalFacadeException("Unimplemented");
  }

  /** Add a member to a group
   *
   * @param group    String group name
   * @param id       String member id
   * @exception CalFacadeException   For invalid usertype values.
   */
  public void addMember(String group, String id) throws CalFacadeException {
    PreparedStatement s = null;

    try {
      ensureAuthUserExists(new UserVO(id));

      s = getPreparedStatement(addGroupMemberSQL);

      s.setString(1, group);
      s.setString(2, id);
      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Remove a member from a group
   *
   * @param group    String group name
   * @param id       String member id
   * @exception CalFacadeException   For invalid usertype values.
   */
  public void removeMember(String group, String id) throws CalFacadeException {
    PreparedStatement s = null;

    try {
      s = getPreparedStatement(deleteGroupMemberSQL);

      s.setString(1, group);
      s.setString(2, id);
      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Delete a group
   *
   * @param  group           AdminGroupVO group object to delete
   * @exception CalFacadeException If there's a problem
   */
  public void removeGroup(AdminGroupVO group) throws CalFacadeException {
    PreparedStatement s = null;

    try {
      s = getPreparedStatement(deleteGroupSQL);

      s.setString(1, group.getName());
      s.executeUpdate();

      s = getPreparedStatement(deleteAllGroupMembersSQL);

      s.setString(1, group.getName());
      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Update a group
   *
   * @param group    AdminGroupVO object
   * @exception CalFacadeException   For invalid usertype values.
   */
  public void updateGroup(AdminGroupVO group) throws CalFacadeException {
    PreparedStatement s = null;

    try {
      /* First ensure owner and event owner exist
       */
      ensureUserExists(group.getGroupOwner());
      ensureUserExists(group.getOwner());

      s = getPreparedStatement(updateGroupSQL);

      s.setString(1, group.getDescription());
      s.setString(2, group.getGroupOwner().getAccount());
      s.setString(3, group.getOwner().getAccount());
      s.setString(4, group.getName());
      s.executeUpdate();
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      close();
    }
  }

  /** Return all admin groups. Never returns null.
   *
   * @return Collection    of AdminGroupVO
   */
  public Collection getAll() throws CalFacadeException {
    return getAdminGroups(null);
  }

  /** Ensure the user exists - create an entry if not
   *
   * @param val      userVO entry
   * @return UserVO  retrieved userVO entry
   */
  private UserVO ensureUserExists(UserVO val) throws CalFacadeException {
    UserVO user = cal.getUser(val.getAccount());

    if (user != null) {
      return user;
    }

    cal.addUser(val);

    return val;
  }

  /** Ensure the authorised user exists - create an entry if not
   *
   * @param val      userVO entry
   * @return UserVO  retrieved userVO entry
   */
  private UserVO ensureAuthUserExists(UserVO val) throws CalFacadeException {
    val = ensureUserExists(val);

    UserAuth uauth = cal.getUserAuth();

    AuthUserVO au = uauth.getUser(val.getAccount());

    if ((au != null) && (au.getUsertype() != UserAuth.noPrivileges)) {
      return val;
    }

    au = new AuthUserVO(val,
                        UserAuth.publicEventUser,
                        "",
                        "",
                        "",
                        "",
                        "");
    uauth.updateUser(au);

    return val;
  }

  /** Return a Collection of group members
   *
   * @param  name             String group name
   * @return Collection       group members
   * @exception CalFacadeException If there's a problem
   */
  private Collection getMembers(String name)
      throws CalFacadeException {
    PreparedStatement s = null;
    ResultSet rs = null;
    Vector v = new Vector();

    try {
      s = getPreparedStatement(findGroupMembersSQL);

      s.setString(1, name);
      rs = s.executeQuery();

      while (rs.next()) {
        UserVO mbr = new UserVO(rs.getString(fldUserid));

        v.addElement(ensureAuthUserExists(mbr));
      }

      if (v.size() == 0) {
        return null;
      }

      return v;
    } catch (Throwable t) {
      throw new CalFacadeException(t);
    } finally {
      if (rs != null) {
        try {
          rs.close();
        } catch (Throwable t) {}
        rs = null;
      }

      close(); // Will free s
      s = null;
    }
  }

  /** Make a populated group object from a result set.
   *
   * @param  rs             ResultSet from search
   * @return AdminGroupVO   grooup object
   */
  private AdminGroupVO makeGroupVO(ResultSet rs) throws Throwable {
    if (!rs.next()) {
      return null;
    }

    AdminGroupVO ug = new AdminGroupVO(
                rs.getString(fldGroupName),
                rs.getString(fldGroupDesc),
                new UserVO(rs.getString(fldGroupGroupOwner)),
                new UserVO(rs.getString(fldGroupOwnerid)),
                null);

    ug.setGroupMembers(getMembers(ug.getName()));

    return ug;
  }
}
