package edu.washington.cac.calfacade.shared;

import java.io.Serializable;
import java.util.Collection;
import java.util.Iterator;

/** A read-only object representing a calendar admin group.
 * Used for web applications to protect the data from the request stream.
 *
 * @author Mike Douglass douglm@rpi.edu
 * @version 2.2
 */
public class AdminGroupROVO implements Serializable {
  protected String name;

  protected String description;

  protected UserVO groupOwner;

  protected UserVO owner;

  /** members of the group
   */
  protected Collection groupMembers;

  /** Only of use to subclasses with set methods
   */
  public AdminGroupROVO() {
  }

  /** Create a new read only object.
   *
   * @param  name          String group name
   * @param  description   String group description
   * @param  groupOwner    UserVO who owns the group.
   * @param  owner         UserVO owner - the owner of events.
   * @param  groupMembers   Collection of UserVO
   */
  public AdminGroupROVO(String name,
                       String description,
                       UserVO groupOwner,
                       UserVO owner,
                       Collection groupMembers) {
    this.name = name;
    this.description = description;
    this.groupOwner = groupOwner;
    this.owner = owner;
    this.groupMembers = groupMembers;
  }

  /** Return the name of the group.
   *
   * @return String        group name
   * @exception CalFacadeException If there's a problem
   */
  public String getName() throws CalFacadeException {
    return name;
  }

  /** Return the description of the group.
   *
   * @return String        group description
   * @exception CalFacadeException If there's a problem
   */
  public String getDescription() throws CalFacadeException {
    return description;
  }

  /** Return the id of the group owner.
   *
   * @return UserVO        group owner
   * @exception CalFacadeException If there's a problem
   */
  public UserVO getGroupOwner() throws CalFacadeException {
    return groupOwner;
  }

  /** Return the ownerid of the group.
   *
   * @return UserVO        group owner
   * @exception CalFacadeException If there's a problem
   */
  public UserVO getOwner() throws CalFacadeException {
    return owner;
  }

  /** Return the members of the group.
   *
   * @return Collection        group members
   * @exception CalFacadeException If there's a problem
   */
  public Collection getGroupMembers() {
    return groupMembers;
  }

  /** Return true if the account name is in the group members.
   */
  public boolean isMember(String account) throws CalFacadeException {
    if (groupMembers == null) {
      return false;
    }

    Iterator it = groupMembers.iterator();

    while (it.hasNext()) {
      UserVO mbr = (UserVO)it.next();

      if (mbr.getAccount().equals(account)) {
        return true;
      }
    }

    return false;
  }
}
