package edu.washington.cac.calfacade.shared;

import java.io.Serializable;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

/** Base object which is closest to the db representation.
 * Possibly used for ejb or hibernate mapping generation.
 *
 * <p>Value object to represent authorised calendar user preferences.
 * These should really be in the same table.
 *
 *  @author Mike Douglass douglm@rpi.edu
 *  @version 1.0
 */
public class AuthUserPrefsVO implements Serializable {
  /* key */
  private int id;

  protected AuthUserVO authUser;

  /** If true automatically add keywords to the preferred list
   */
  protected boolean autoAddKeywords = true;

  /** Users preferred keywords.
   */
  protected Set preferredKeywords;

  /** If true automatically add locations to the preferred list
   */
  protected boolean autoAddLocations = true;

  /** Users preferred locations.
   */
  protected Set preferredLocations;

  /** If true automatically add sponsors to the preferred list
   */
  protected boolean autoAddSponsors = true;

  /** Users preferred sponsors.
   */
  protected Set preferredSponsors;

  /* ====================================================================
   *                   Bean methods
   * ==================================================================== */

  public void setId(int val) {
    id = val;
  }

  public int getId() {
    return id;
  }

  public void setAuthUser(AuthUserVO val) {
    authUser = val;
  }

  public AuthUserVO getAuthUser() {
    return authUser;
  }

  /** If true automatically add keywords to the preferred list
   */
  public void setAutoAddKeywords(boolean val) {
    autoAddKeywords = val;
  }

  public boolean getAutoAddKeywords() {
    return autoAddKeywords;
  }

  public void setPreferredKeywords(Set val) {
    preferredKeywords = val;
  }

  public Set getPreferredKeywords() {
    return preferredKeywords;
  }

  /** If true automatically add locations to the preferred list
   */
  public void setAutoAddLocations(boolean val) {
    autoAddLocations = val;
  }

  public boolean getAutoAddLocations() {
    return autoAddLocations;
  }

  public void setPreferredLocations(Set val) {
    preferredLocations = val;
  }

  public Set getPreferredLocations() {
    return preferredLocations;
  }

  /** If true automatically add sponsors to the preferred list
   */
  public void setAutoAddSponsors(boolean val) {
    autoAddSponsors = val;
  }

  public boolean getAutoAddSponsors() {
    return autoAddSponsors;
  }

  public void setPreferredSponsors(Set val) {
    preferredSponsors = val;
  }

  public Set getPreferredSponsors() {
    return preferredSponsors;
  }

  /* ====================================================================
   *                   Convenience methods
   * ==================================================================== */

  /** Add the keyword to the preferred keywords. Return true if it was
   * added, false if it was in the list
   *
   * @param k          Keyword to add
   * @return boolean   true if added
   */
  public boolean addKeyword(KeywordVO k) {
    if (preferredKeywords == null) {
      preferredKeywords = new TreeSet();
    } else if (preferredKeywords.contains(k)) {
      return false;
    }

    preferredKeywords.add(k);
    return true;
  }

  /** Remove the keyword from the preferred keywords. Return true if it was
   * removed, false if it was not in the list
   *
   * @param k          Keyword to remove
   * @return boolean   true if removed
   */
  public boolean removeKeyword(KeywordVO k) {
    if (preferredKeywords == null) {
      return false;
    }

    if (!preferredKeywords.contains(k)) {
      return false;
    }

    preferredKeywords.remove(k);

    return true;
  }

  /** Add the location to the preferred keywords. Return true if it was
   * added, false if it was in the list
   *
   * @param l          LocationVO to add
   * @return boolean   true if added
   */
  public boolean addLocation(LocationVO l) {
    if (preferredLocations == null) {
      preferredLocations = new TreeSet();
    } else if (preferredLocations.contains(l)) {
      return false;
    }

    preferredLocations.add(l);
    return true;
  }

  /** Remove the location from the preferred locations. Return true if it was
   * removed, false if it was not in the list
   *
   * @param l          LocationVO to remove
   * @return boolean   true if removed
   */
  public boolean removeLocation(LocationVO l) {
    if (preferredLocations == null) {
      return false;
    }

    if (!preferredLocations.contains(l)) {
      return false;
    }

    preferredLocations.remove(l);

    return true;
  }

  /** Add the sponsor to the preferred sponsors. Return true if it was
   * added, false if it was in the list
   *
   * @param s          SponsorVO to add
   * @return boolean   true if added
   */
  public boolean addSponsor(SponsorVO s) {
    if (preferredSponsors == null) {
      preferredSponsors = new TreeSet();
    } else if (preferredSponsors.contains(s)) {
      return false;
    }

    preferredSponsors.add(s);
    return true;
  }

  /** Remove the sponsor from the preferred sponsors. Return true if it was
   * removed, false if it was not in the list
   *
   * @param s          SponsorVO to remove
   * @return boolean   true if removed
   */
  public boolean removeSponsor(SponsorVO s) {
    if (preferredSponsors == null) {
      return false;
    }

    if (!preferredSponsors.contains(s)) {
      return false;
    }

    preferredSponsors.remove(s);

    return true;
  }

  /* ====================================================================
   *                   Object methods
   * ==================================================================== */

  public String toString() {
    StringBuffer sb = new StringBuffer();

    sb.append("AuthUserPrefsVO{id=");
    sb.append(id);
    sb.append(", autoAddKeywords=");
    sb.append(autoAddKeywords);
    sb.append(", autoAddLocations=");
    sb.append(autoAddLocations);
    sb.append(", autoAddSponsors=");
    sb.append(autoAddSponsors);
    sb.append("}");

    return sb.toString();
  }

  public Object clone() {
    AuthUserPrefsVO aup = new AuthUserPrefsVO();

    aup.setId(getId());

    // Don't clone the authuser - sets up a loop

    aup.setAutoAddKeywords(getAutoAddKeywords());

    TreeSet nts = new TreeSet();
    Set ts = getPreferredKeywords();

    if (ts != null) {
      Iterator it = ts.iterator();

      while (it.hasNext()) {
        nts.add((KeywordVO)((KeywordVO)it.next()).clone());
      }
    }

    aup.setPreferredKeywords(nts);

    aup.setAutoAddLocations(getAutoAddLocations());

    nts = new TreeSet();
    ts = getPreferredLocations();

    if (ts != null) {
      Iterator it = ts.iterator();

      while (it.hasNext()) {
        nts.add((LocationVO)((LocationVO)it.next()).clone());
      }
    }

    aup.setPreferredLocations(nts);

    aup.setAutoAddSponsors(getAutoAddSponsors());

    nts = new TreeSet();
    ts = getPreferredSponsors();

    if (ts != null) {
      Iterator it = ts.iterator();

      while (it.hasNext()) {
        nts.add((SponsorVO)((SponsorVO)it.next()).clone());
      }
    }

    aup.setPreferredSponsors(nts);

    return aup;
  }
}
