package edu.washington.cac.calfacade.shared;

import edu.washington.cac.calfacade.shared.filter.FilterVO;

import java.io.Serializable;
import java.sql.Timestamp;
import java.util.Collection;
import java.util.TreeSet;

/** This class acts as a low level interface to the calendar database.
 *
 * <p>This interface provides a view of the data as seen by the supplied user
 * id. This may or may not be the actual authenticated user of whatever
 * application is driving it.
 *
 * <p>This is of particular use for public events administration. A given
 * authenticated user may be the member of a number of groups, and this module
 * will be initialised with the id of one of those groups. At some point
 * the authenticated user may choose to switch identities to manage a different
 * group.
 *
 * <p>The UserAuth object returned by getUserAuth usually represents the
 * authenticated user and determines the rights that user has.
 *
 * @author Mike Douglass   douglm@rpi.edu
 */
public interface Calintf extends Serializable {
  /** Must be called to initialise the new object.
   *
   * @param authUser    String authenticated user of the application
   * @param rights      int access rights value as defined by
   *                    edu.washington.cac.calfacade.shared.UserAuth
   * @param user        String user we are acting as. If null we use authUser
   * @param isGuest     boolean true if user is guest
   * @param publicAdmin boolean true if this is a public events admin app
   * @param synchId     non-null if this is for synchronization. Identifies the
   *                    client end.
   * @param debug       boolean true to turn on debugging trace
   */
  public void init(String authUser,
                   int rights,
                   String user,
                   boolean isGuest,
                   boolean publicAdmin,
                   String synchId,
                   boolean debug) throws CalFacadeException;

  /** Switch to the given non-null user. The access rights are determined by
   * the previous call to init.
   *
   * @param val         String user id
   */
  public void setUser(String val) throws CalFacadeException;

  /** Signal the start of a sequence of operations. These overlap transactions
   * in that there may be 0 to many transactions started and ended within an
   * open/close call and many open/close calls within a transaction.
   *
   * <p>The open close methods are mainly associated with web style requests
   * and open will usually be called early in the incoming request handling
   * and close will always be called on the way out allowing the interface an
   * opportunity to reacquire (on open) and release (on close) any resources
   * such as connections.
   */
  public void open() throws CalFacadeException;

  /** Call on the way out after handling a request..
   */
  public void close() throws CalFacadeException;

  /** Start a (possibly long-running) transaction. In the web environment
   * this might do nothing. The endTransaction method should in some way
   * check version numbers to detect concurrent updates and fail with an
   * exception.
   */
  public void beginTransaction() throws CalFacadeException;

  /** End a (possibly long-running) transaction. In the web environment
   * this should in some way check version numbers to detect concurrent updates
   * and fail with an exception.
   */
  public void endTransaction() throws CalFacadeException;

  /** Call if there has been an error during an update process.
   */
  public void rollbackTransaction() throws CalFacadeException;

  /** Returns a value object representing the current user.
   *
   * @return UserVO       representing the current user
   */
  public UserVO getUserVO() throws CalFacadeException;

  /** Add a user to the database
   *
   * @param user   UserVO object to add.
   */
  public void addUser(UserVO user) throws CalFacadeException;

  /** Returns a value object representing the user with the given id.
   *
   * @param id            int id of user entry
   * @return UserVO       representing the current user
   */
  public UserVO getUser(int id) throws CalFacadeException;

  /** Returns a value object representing the user with the given account name.
   *
   * @param account       string account name of user entry
   * @return UserVO       representing the current user
   */
  public UserVO getUser(String account) throws CalFacadeException;

  /** Get a UserAuth object which allows the application to determine what
   * special rights the user has.
   *
   * <p>Note that the returned object may require a one time initialisation
   * call to set it up correctly.
   * @see edu.washington.cac.calfacade.shared.UserAuth#initialise(String, int)
   * @see edu.washington.cac.calfacade.shared.UserAuth#initialise(String, Object)
   *
   * @return UserAuth    implementation.
   */
  public UserAuth getUserAuth() throws CalFacadeException;

  /** Get an AdminGroups object which allows the application to handle
   * administrative groups.
   *
   * @return AdminGroups    implementation.
   */
  public AdminGroups getAdminGroups() throws CalFacadeException;

  /** Refresh the users events cache
   */
  public void refreshEvents() throws CalFacadeException;

  /* ====================================================================
   *                   Global parameters
   * ==================================================================== */

  /** Get the value of the current public events timestamp.
   *
   * @return long       current timestamp value
   */
  public long getPublicLastmod() throws CalFacadeException;

  /** Get a name uniquely.identifying this system. This should take the form <br/>
   *   name@host
   * <br/>where<ul>
   * <li>name identifies the particular calendar system at the site</li>
   * <li>host part identifies the domain of the site.</li>..
   * </ul>
   *
   * @return String    globally unique system identifier.
   */
  public String getSysid() throws CalFacadeException;

  /* ====================================================================
   *                   Calendars and search
   * ==================================================================== */

  /** Return a list of calendar we can present to the client as a list of
   * available calendars.
   *
   * @return Collection   of FilterVO sorted by name
   */
  public Collection getCalendars() throws CalFacadeException;

  /** Set a search filter using the suppplied search string
   *
   * @param val    String search parameters
   */
  public void setSearch(String val) throws CalFacadeException;

  /** Return the current search string.
   *
   * @return  String     search parameters
   */
  public String getSearch() throws CalFacadeException;

  /** Set the calendar we are interested in. This is represented by the id
   * of a calendar.
   *
   * @param  val     id of calendar - 0 for none
   * @return boolean true for OK, false for unknown calendar
   */
  public boolean setCalendar(int val) throws CalFacadeException;

  /** Get the name of the current calendar we are displaying
   *
   * @return String    name of calendar or null for all events
   */
  public String getCalendarTitle() throws CalFacadeException;

  /** This method is called to add a calendar. No consistency checks are
   * carried out at the moment.
   *
   * @param val  String values for db
   */
  public void addCalendarDef(String val) throws CalFacadeException;

  /** Add a filter to the database. All references must be set.
   */
  public void addFilter(FilterVO val) throws CalFacadeException;

  /** Update a filter
   *
   * @param  val           FilterVO object to upate
   * @exception CalFacadeException If there's a problem
   */
  public void updateFilter(FilterVO val) throws CalFacadeException;

  /* * Set a filter which returns events starting after a given Timestamp.
   *
   * @param val   Timestamp object.
   * /
  public void setLastmodGECalendar(Timestamp val) throws CalFacadeException;

  /* * Set a filter which returns events starting after a given date of the
   * form yyyymmdd.
   *
   * @param date   String of form yyyymmdd.
   * /
  public void setLastmodGEDateCalendar(String date) throws CalFacadeException;

  /* * Set a filter which returns events starting after a given date and time
   * of the form yyyymmddhhmm.
   *
   * @param dateTime   String of form yyyymmddhhmm.
   * /
  public void setLastmodGEDateTimeCalendar(String dateTime)
      throws CalFacadeException;
  */

  /* ====================================================================
   *                   Subscriptions
   * ==================================================================== */

  /** Ensure that the users subscriptions match a set of calendars
   *
   * @param crs      TreeSet of Integer calendar ids
   * @return int[]   Number of calendars subscribed to, and number of
   *                 calendars unsubscribed to
   */
  public int[] fixSubscriptions(TreeSet crs) throws CalFacadeException;

  /** Determine if user is subscribed to a calendar
   *
   * @param id       int calendar id
   * @return boolean true if subscribed to calendar
   */
  public boolean getSubscribed(int id) throws CalFacadeException;

  /* * Get the subscriptions for the user
   *
   * @param user        UserVO object
   * /
  public void getSubscriptions(UserVO user) throws CalFacadeException;*/

  /* ====================================================================
   *                   Keywords
   * ==================================================================== */

  /** Return all keywords for this user. They will have their attributes set.
   * For a public admin client this is the same as getPublicKeywords().
   *
   * @return Collection  of keywords value objects
   */
  public Collection getKeywords() throws CalFacadeException;

  /** Add a Keyword to the database. The id will be set in the parameter
   * object.
   *
   * @param val   KeywordVO object to be added
   */
  public void addKeyword(KeywordVO val) throws CalFacadeException;

  /** Add a Keyword to the database and give it the assigned key.
   * Only for restores
   *
   * @param val   KeywordVO object to be added with key set
   */
  public void addKeywordWithKey(KeywordVO val) throws CalFacadeException;

  /** Return a keyword with the given id
   *
   * @param id     int id of the keyword
   * @return KeywordVo object representing the keyword in question
   *                     null if it doesn't exist.
   */
  public KeywordVO getKeyword(int id) throws CalFacadeException;

  /* ====================================================================
   *                   Public Keywords
   * ==================================================================== */

  /** Return all public keywords
   *
   * @return Collection  of keywords value objects
   */
  public Collection getPublicKeywords() throws CalFacadeException;

  /** Return given public keyword
   *
   * @param id     int id of the keyword
   * @return KeywordVO[]  array of keywords value objects
   */
  public KeywordVO getPublicKeyword(int id) throws CalFacadeException;

  /** Replace a keyword in the database.
   *
   * @param val   KeywordVO object to be replaced
   */
  public void replaceKeyword(KeywordVO val) throws CalFacadeException;

  /** Delete the given keyword
   *
   * @param val      KeywordVO object to be deleted
   * @return boolean false if it didn't exist, true if it was deleted.
   */
  public boolean deleteKeyword(KeywordVO val) throws CalFacadeException;

  /** Check to see if a Keyword is referenced.
   *
   * @param val      KeywordVO object to check
   * @return boolean true if the Keyword is referenced somewhere
   */
  public boolean checkKeywordRefs(KeywordVO val) throws CalFacadeException;

  /** ===================================================================
   *                   Keyword attributes
   *  =================================================================== */

  /* * Get the keyword attributes entry for a keyword id
   *
   * @param keywordid       int id of the keyword entry
   * @return Set            of keyword attributes
   * /
  public Set getKeywordAttrs(int keywordid) throws CalFacadeException;*/

  /** Add a keyword attribute name/value pair for a keyword
   *
   * @param val             KeywordVO object to update
   * @param name            String attribute name
   * @param aval            String attribute value
   */
  public void addKeywordAttr(KeywordVO val,
                             String name,
                             String aval) throws CalFacadeException;

  /** Remove a keyword attribute name/value pair for a keyword
   *
   * @param val             KeywordVO object to update
   * @param name            String attribute name
   * @param aval            String attribute value
   */
  public void removeKeywordAttr(KeywordVO val,
                                String name,
                                String aval) throws CalFacadeException;

  /** ===================================================================
   *                   Locations
   *  =================================================================== */

  /** Return all locations for this user. Returns an empty collection for
   * no locations. If this is a public admin instance, this will return all
   * public locations owned by the current user. Otherwise it returns all
   * personal locations owned by the user.
   *
   * @return Collection        of location value objects
   */
  public Collection getLocations() throws CalFacadeException;

  /** Add a location to the database. The id will be set in the parameter
   * object.
   *
   * @param val   LocationVO object to be added
   */
  public void addLocation(LocationVO val) throws CalFacadeException;

  /** Replace a location in the database.
   *
   * @param val   LocationVO object to be replaced
   */
  public void replaceLocation(LocationVO val) throws CalFacadeException;

  /** Return a location with the given id
   *
   * @param id     int id of the location
   * @return LocationVO object representing the location in question
   *                     null if it doesn't exist.
   */
  public LocationVO getLocation(int id) throws CalFacadeException;

  /** Delete a location
   *
   * @param val      LocationVO object to be deleted
   * @return boolean false if it din't exist, true if it was deleted.
   */
  public boolean deleteLocation(LocationVO val) throws CalFacadeException;

  /** Check to see if a location is referenced.
   *
   * @param val      LocationVO object to be checked
   * @return boolean true if the location is referenced somewhere
   */
  public boolean checkLocationRefs(LocationVO val) throws CalFacadeException;

  /** ===================================================================
   *                   Public Locations
   *  =================================================================== */

  /** Return public locations - excluding reserved entries. Returns an empty
   * collection for no locations.
   *
   * @return Collection        of location value objects
   */
  public Collection getPublicLocations() throws CalFacadeException;

  /** Return public locations - optionally including reserved entries.
   * Returns an empty collection for no locations.
   *
   * @param includeReserved boolean true include reserved, false exclude
   * @return Collection        of location value objects
   */
  public Collection getPublicLocations(boolean includeReserved)
      throws CalFacadeException;

  /** Return given public location
   *
   * @param id          int id of the location
   * @return LocationVo object representing the location in question
   *                     null if it doesn't exist.
   */
  public LocationVO getPublicLocation(int id) throws CalFacadeException;

  /* * Delete given public location
   *
   * @param val      LocationVO object to be deleted
   * @return boolean false if it didn't exist, true if it was deleted.
   * /
  public boolean deletePublicLocation(LocationVO val) throws CalFacadeException;
  */

  /** ===================================================================
   *                   Sponsors
   *  =================================================================== */

  /** Return all sponsors for this user. Returns an empty collection for
   * no sponsors. If this is a public admin instance, this will return all
   * public sponsors owned by the current user. Otherwise it returns all
   * personal sponsors owned by the user.
   *
   * <p>(Note the current implementation does not allow forthe creation of
   * personal sponsors.
   *
   * @return Collection     of sponsors
   */
  public Collection getSponsors() throws CalFacadeException;

  /** Return a sponsor with the given id
   *
   * @param id     int id of the sponsor
   * @return SponsorVo object representing the sponsor in question
   *                     null if it doesn't exist.
   */
  public SponsorVO getSponsor(int id) throws CalFacadeException;

  /** Add a sponsor to the database. The id will be set in the parameter
   * object.
   *
   * @param val   SponsorVO object to be added
   */
  public void addSponsor(SponsorVO val) throws CalFacadeException;

  /** Replace a sponsor in the database.
   *
   * @param val   SponsorVO object to be replaced
   */
  public void replaceSponsor(SponsorVO val) throws CalFacadeException;

  /** Delete a sponsor
   *
   * @param val      SponsorVO object to be deleted
   * @return boolean false if it didn't exist, true if it was deleted.
   */
  public boolean deleteSponsor(SponsorVO val) throws CalFacadeException;

  /** Check to see if a sponsor is referenced.
   *
   * @param val      SponsorVO object to be checked
   * @return boolean true if the sponsor is referenced somewhere
   */
  public boolean checkSponsorRefs(SponsorVO val) throws CalFacadeException;

  /** ===================================================================
   *                   Public Sponsors
   *  =================================================================== */

  /** Return public sponsors - excluding reserved entries. Returns an empty
   * collection for no sponsors.
   *
   * @return Collection        of sponsor value objects
   */
  public Collection getPublicSponsors() throws CalFacadeException;

  /** Return all public sponsors - optionally including reserved entries.
   * Returns an empty collection for no sponsors.
   *
   * @param includeReserved boolean true include reserved, false exclude
   * @return Collection     of sponsors
   */
  public Collection getPublicSponsors(boolean includeReserved)
      throws CalFacadeException;

  /** Return given public sponsor
   *
   * @param id         int id of the sponsor
   * @return SponsorVo object representing the sponsor in question
   *                     null if it doesn't exist.
   */
  public SponsorVO getPublicSponsor(int id) throws CalFacadeException;

  /** ===================================================================
   *                   Events
   *  =================================================================== */

  /** Return a single event for the current user
   *
   * @param   eventId   int id of the event
   * @return  EventVO   value object representing event.
   */
  public EventVO getEvent(int eventId) throws CalFacadeException;

  /** Return filtered events for the current user.
   *
   * @return Collection of event value objects
   */
  public Collection getEvents() throws CalFacadeException;

  /** Return the events for the given day as an array of value objects
   *
   * @param   date        java.util.Date object defining day
   * @return  Collection  one days events, empty for no events.
   */
  public Collection getDaysEvents(java.util.Date date) throws CalFacadeException;

  /** Add an event to the database. The id and uid will be set in the parameter
   * object.
   *
   * @param val   EventVO object to be added
   */
  public void addEvent(EventVO val) throws CalFacadeException;

  /** Update an event in the database.
   *
   * <p>This method will set any synchronization state entries to modified
   * unless we are synchronizing in which case that belonging to the current
   * user is set to mark the event as synchronized
   *
   * @param val   EventVO object to be replaced
   * @exception CalFacadeException If there's a db problem or problem with
   *     the event
   */
  public void updateEvent(EventVO val) throws CalFacadeException;

  /** Fix up the event keywords after an update.
   *
   * @param val         updated EventVO object
   * @param oldEvent    EventVO object before changes applied. null for a
   *                    new event.
   */
  public void fixKeywords(EventVO val,
                          EventVO oldEvent) throws CalFacadeException;

  /** Delete an event
   *
   * @param val      EventVO object to be deleted
   * @return boolean false if it didn't exist, true if it was deleted.
   * @exception CalFacadeException If there's a database access problem
   */
  public boolean deleteEvent(EventVO val) throws CalFacadeException;

  /** Assign a guid to an event which must exist in the db. A noop if this
   * event already has a guid.
   *
   * @param val      EventVO object
   */
  public void assignGuid(EventVO val) throws CalFacadeException;

  /** ===================================================================
   *                   Public Events
   *  =================================================================== */

  /** Return the public events for the current user. This is an administrative
   * function for the public events admin client.
   *
   * @return Collection  populated event value objects
   */
  public Collection findPublicEvents() throws CalFacadeException;

  /** Return the public events for the current user within the given date
   * range. This is an administrative function for the public events admin
   * client.
   *
   * @param startDate    java.util.Date start date - may be null
   * @param endDate      java.util.Date end of period - may be null.
   * @return Collection  populated event value objects
   */
  public Collection findPublicEvents(java.util.Date startDate,
                                     java.util.Date endDate)
          throws CalFacadeException;

  /* * Return the public events for the given day as an array of value objects
   *
   * @param   date    MyCalendar object defining day
   * @return  EventVO[]  one days events or null for no events.
   * /
  public EventVO[] getDaysPublicEvents(MyCalendarVO date)
      throws CalFacadeException;*/

  /* ====================================================================
   *                   EventRefs
   * ==================================================================== */

  /** Create an event ref for the current user.
   *
   * @param  event      EventVO object
   * @return  boolean   false for no such item.
   */
  public boolean addEventRef(EventVO event) throws CalFacadeException;

  /** Delete an event ref for the current user.
   *
   * @param  event          EventVO object
   * @return  boolean       false for no such item.
   */
  public boolean deleteEventRef(EventVO event) throws CalFacadeException;

  /** Delete a subscribed event for the current user.
   *
   * <p>This is a partially implemented feature which currently only changes
   * the synch state. What we need is in effect a reverse event ref - I guess
   * a new purpose, which indicates the event is hidden for this user.
   *
   * @param  event          EventVO object
   * @return  boolean       false for no such item.
   */
  public boolean deleteSubscribedEvent(EventVO event) throws CalFacadeException;

  /* ====================================================================
   *                   Synchronization
   * ==================================================================== */

  /** Get the synch info for the current device and user.
   *
   * @return SynchInfoVO object or null.
   */
   public SynchInfoVO getSynchInfo() throws CalFacadeException;

  /** Add the synch info for the current device and user.
   *
   * @param val    SynchInfoVO object to add
   */
  public void addSynchInfo(SynchInfoVO val) throws CalFacadeException;

  /** Update the synch info for the current device and user. There must be
   * exactly one object to update.
   *
   * @param val    SynchInfoVO object to update
   */
  public void updateSynchInfo(SynchInfoVO val)
      throws CalFacadeException;

  /** Return synchronization state for the given event in the current
   * synchronization context as defined by the initialisation parameters,
   * that is the user and synchId. If the event has never been involved in a
   * synch process there will be no related synch state object.
   *
   * @param ev            EventVO object .
   * @return SynchStateVO object or null.
   */
  public SynchStateVO getSynchState(EventVO ev)
      throws CalFacadeException;

  /** Get the deleted synch states for the current synch.
   *
   * @returns Collection of synch state objects (possibly empty)
   */
  public Collection getDeletedSynchStates() throws CalFacadeException;

  /** Add the synch state for the given event and user.
   *
   * @param val      SynchStateVO object.
   */
  public void addSynchState(SynchStateVO val)
      throws CalFacadeException;

  /** Update the synch state for the given event and user. There must be
   * exactly one object to update.
   *
   * @param val    SynchStateVO object with associated event and user
   */
  public void updateSynchState(SynchStateVO val)
      throws CalFacadeException;

  /** Update the synch state for the given event. There may be many objects
   * to update.
   *
   * <p>In general it only makes sense to call this method to set the state
   * to modified or deleted. A modified event is modified for any user who
   * has access to it. On the other hand, an event can only be synchronized
   * for a single user.
   *
   * @param ev     EventVO representing associated event
   * @param state  int new state - defined in SynchStateVO
   * @return int   number of updated entries.
   */
  public int setSynchState(EventVO ev, int state)
        throws CalFacadeException;

  /** Get the synch data associated with the given object..
   *
   * @param val    SynchStateVO object
   */
  public void getSynchData(SynchStateVO val) throws CalFacadeException;

  /** Set the synch state and data for the given user and event.
   * There may be zero to many objects to update.
   *
   * @param val    SynchStateVO object with associated event and user
   */
  public void setSynchData(SynchStateVO val) throws CalFacadeException;

  /** Update the synch states for the current user and device.
   *
   * <p>This is called at the termination of synchronization to set any state
   * objects appropriately.
   *
   * <p><table cols="2">
   * <th><td>Old state</td><td>New state</td></th>
   * <tr><td>UNKNOWN</td><td>SYNCHRONIZED</td></tr>
   * <tr><td>SYNCHRONIZED</td><td>SYNCHRONIZED</td></tr>
   * <tr><td>NEW</td><td>SYNCHRONIZED</td></tr>
   * <tr><td>MODIFIED</td><td>SYNCHRONIZED</td></tr>
   * <tr><td>DELETED</td><td>entry deleted</td></tr>
   * <tr><td>CLIENT_DELETED</td><td>unchanged</td></tr>
   * <tr><td>CLIENT_DELETED_UNDELIVERED</td><td>CLIENT_DELETED</td></tr>
   * <tr><td>CLIENT_MODIFIED</td><td>unchanged</td></tr>
   * <tr><td>CLIENT_MODIFIED_UNDELIVERED</td><td>CLIENT_MODIFIED</td></tr>
   * </table>
   */
  public void updateSynchStates() throws CalFacadeException;

  /* ====================================================================
   *                       Recurrences
   * This is currently under development at RPI
   * ==================================================================== */

  /** Add a recurrence to the database.
   *
   * @param val      RecurVO object.
   */
  public void addRecur(RecurVO val) throws CalFacadeException;

  /** Get a recurrence associated with an event.
   *
   * @param ev      Associated EventVO
   * @return RecurVO  recurrence object or null
   */
  public RecurVO getRecur(EventVO ev) throws CalFacadeException;

  /** Remove a recurrence associated with an event.
   *
   * @param ev      Associated EventVO
   */
  public void deleteRecur(EventVO ev) throws CalFacadeException;

  /** Update a recurrence to the database.
   *
   * @param val      RecurVO object.
   */
  public void updateRecur(RecurVO val) throws CalFacadeException;
}

