package edu.washington.cac.calfacade.shared;

import java.io.Serializable;

/** Class representing a guid. Used in icalendar objects
 *
 * @author Mike Douglass   douglm@rpi.edu
 *  @version 1.0
 */
public class GuidVO implements Serializable {
  /* If this is a uwcal generated guid we will parse it into the following fields.
   */
  /** Id of the related object
   */
  private int id = CalFacadeDefs.unsavedItemKey;

  private String sysid;

  /* A non uwcal guid will be placed here
   */
  private String externalGuid;

  /** Expect this prefixing the sysid
   */
  private static final String uidid = "uwcal";

  public GuidVO() {
  }

  /** Create a guid by specifying all its fields

      @param id          Unique ID for the keyword
      @param sysid       String system id
    */
  public GuidVO(int id,
               String sysid) {
    this.id = id;
    this.sysid = sysid;
  }

  /** Create an external guid

      @param externalGuid
    */
  public GuidVO(String externalGuid) {
    id = CalFacadeDefs.unsavedItemKey;
    sysid = null;
    this.externalGuid = externalGuid;
  }

  /**
    Set the id for this keyword
    @param val    int Keyword id
   */
  public void setId(int val) {
    id = val;
  }

  /** Get the kwywords unique id
   *
   * @return int    the keywords unique id
   */
  public int getId() {
    return id;
  }

  /** Set the sysid
   *
   * @param val    String sysid
   */
  public void setSysid(String val) {
    sysid = val;
  }

  /** Get the sysid
   *
   * @return String   sysid
   */
  public String getSysid() {
    return sysid;
  }

  /** Set the externalGuid. Resets internal fields.
   *
   * @param val    String externalGuid
   */
  public void setExternalGuid(String val) {
    id = CalFacadeDefs.unsavedItemKey;
    sysid = null;
    externalGuid = val;
  }

  /** Get the externalGuid
   *
   * @return String   externalGuid
   */
  public String getExternalGuid() {
    return externalGuid;
  }

  /* ====================================================================
   *                        Factory methods
   * ==================================================================== */

  /** This method attempts to decompose the guid and check its validity.
   * Returns an object with fields set from the value.
   *
   * <p>externalGuid will be set if the value cannot be parsed.
   *
   * @param val      String representation of the guid
   * @return GuidVO  representing the decomposed value.
   */
  public static GuidVO valueOf(String val) throws CalFacadeException {
    if (val == null) {
      throw new CalFacadeException("Null value");
    }

    int endId = val.indexOf("-");

    if (endId < 0) {
      return new GuidVO(val);
    }

    String possibleId = val.substring(0, endId);
    int id;

    try {
      id = Integer.valueOf(possibleId).intValue();
    } catch (Throwable t) {
      return new GuidVO(val);
    }

    if (endId == val.length()) {
      return new GuidVO(val);
    }

    endId++;
    if (val.indexOf(uidid, endId) != endId) {
      return new GuidVO(val);
    }

    endId += uidid.length();

    if (val.indexOf("-", endId) != endId) {
      return new GuidVO(val);
    }

    endId++;

    return new GuidVO(id, val.substring(endId));
  }

  /* ====================================================================
   *                        Object methods
   * ==================================================================== */

  public String toString() {
    StringBuffer sb = new StringBuffer();

    sb.append(id);
    sb.append("-");
    sb.append(uidid);
    sb.append("-");
    sb.append(sysid);

    return sb.toString();
  }
}

