package edu.washington.cac.calfacade.shared;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Vector;

import net.fortuna.ical4j.util.DateTimeFormat;

/**
 * Parsed recurrence rule information for an event.
 *
 * @author Mike Douglass douglm@rpi.edu
 * @version 1.0
 */
public class RecurRuleVO implements Comparable, Serializable {
  private int id;

  /** Associated RecurVO */
  private RecurVO recur;

  /** true for a recurrence, false for an exclusion */
  private boolean recurrence;

  /* Wait for java 5 */
  public static final int freqSecondly = 0;
  public static final int freqMinutely = 1;
  public static final int freqHourly = 2;
  public static final int freqDaily = 3;
  public static final int freqWeekly = 4;
  public static final int freqMonthly = 5;
  public static final int freqYearly = 6;

  public static final String[] freqNames = {
    "SECONDLY",
    "MINUTELY",
    "HOURLY",
    "DAILY",
    "WEEKLY",
    "MONTHLY",
    "YEARLY"};

  private int freq;

  /** Only one of until or count */
  private Date until;
  private int count;

  /** Each of the following may appear only once.
   */

  private int interval = -1;  // < 0 means absent

  private IntArrayList bySecond;
  private IntArrayList byMinute;
  private IntArrayList byHour;

  /** Each element is a 2 element integer array. First element is the ordwk
      value, second is a week day number, Sunday = 0
   * /
  private ArrayList byDay;
  */

  /* For the moment ignore ordwk */
  private IntArrayList byDay;

  private IntArrayList byMonthday;
  private IntArrayList byYearday;
  private IntArrayList byWeekno;
  private IntArrayList byMonth;
  private IntArrayList bySetpos;

  private int wkst = -1;  // If set 0 = Sunday, 6 = Saturday

  private final static String[] weekday = {
    "SU",
    "MO",
    "TU",
    "WE",
    "TH",
    "FR",
    "SA"
  };

  /** If we derived this object from an rfc2445 rule we store the text here
   * so we can replay it on synchronization. This attempts to avoid
   * differing representations of the same rule from triggering needless
   * updates.
   */
  private String ruleText;

  /** Constructor
   *
   */
  public RecurRuleVO() {
  }

  public void setId(int val) {
    id = val;
  }

  public int getId() {
    return id;
  }

  public void setRecur(RecurVO val) {
    recur = val;
  }

  public RecurVO getRecur() {
    return recur;
  }

  public void setRecurrence(boolean val) {
    recurrence = val;
  }

  public boolean getRecurrence() {
    return recurrence;
  }

  public void setFreq(int val) {
    freq = val;
  }

  public int getFreq() {
    return freq;
  }

  public void setUntil(Date val) {
    until = val;
  }

  public Date getUntil() {
    return until;
  }

  public void setCount(int val) {
    count = val;
  }

  public int getCount() {
    return count;
  }

  public void setInterval(int val) {
    interval = val;
  }

  public int getInterval() {
    return interval;
  }

  public void setBySecond(IntArrayList val) {
    bySecond = val;
  }

  public IntArrayList getBySecond() {
    return bySecond;
  }

  public void setByMinute(IntArrayList val) {
    byMinute = val;
  }

  public IntArrayList getByMinute() {
    return byMinute;
  }

  public void setByHour(IntArrayList val) {
    byHour = val;
  }

  public IntArrayList getByHour() {
    return byHour;
  }

  /** This needs changing ordwk is not implemented
   */
  public void setByDay(IntArrayList val) {
    byDay = val;
  }

  public IntArrayList getByDay() {
    return byDay;
  }

  public void setByMonthday(IntArrayList val) {
    byMonthday = val;
  }

  public IntArrayList getByMonthday() {
    return byMonthday;
  }

  public void setByYearday(IntArrayList val) {
    byYearday = val;
  }

  public IntArrayList getByYearday() {
    return byYearday;
  }

  public void setByWeekno(IntArrayList val) {
    byWeekno = val;
  }

  public IntArrayList getByWeekno() {
    return byWeekno;
  }

  public void setByMonth(IntArrayList val) {
    byMonth = val;
  }

  public IntArrayList getByMonth() {
    return byMonth;
  }

  public void setBySetpos(IntArrayList val) {
    bySetpos = val;
  }

  public IntArrayList getBySetpos() {
    return bySetpos;
  }

  public void setWkst(int val) {
    wkst = val;
  }

  public int getWkst() {
    return wkst;
  }

  public void setRuleText(String val) {
    ruleText = val;
  }

  public String getRuleText() {
    if (ruleText != null) {
      return ruleText;
    }

    StringBuffer sb = new StringBuffer();

    sb.append("FREQ=");
    sb.append(freqNames[freq]);

    if (until != null) {
      DateTimeFormat dtf = DateTimeFormat.getInstance();
      sb.append("; UNTIL=");
      sb.append(dtf.format(until));
    } else if (count != 0) {
      sb.append("; COUNT=");
      sb.append(count);
    }

    if (interval >= 0) {
      sb.append("; INTERVAL=");
      sb.append(interval);
    }

    if (bySecond != null) {
      sb.append("; BYSECOND=");
      sb.append(bySecond.encode());
    }

    if (byMinute != null) {
      sb.append("; BYMINUTE=");
      sb.append(byMinute.encode());
    }

    if (byHour != null) {
      sb.append("; BYHOUR=");
      sb.append(byHour.encode());
    }

    if (byDay != null) {
      sb.append("; BYDAY=");
      sb.append(byDay.encode());
    }

    if (byMonthday != null) {
      sb.append("; BYMONTHDAY=");
      sb.append(byMonthday.encode());
    }

    if (byYearday != null) {
      sb.append("; BYYEARDAY=");
      sb.append(byYearday.encode());
    }

    if (byWeekno != null) {
      sb.append("; BYWEEKNO=");
      sb.append(byWeekno.encode());
    }

    if (byMonth != null) {
      sb.append("; BYMONTH=");
      sb.append(byMonth.encode());
    }

    if (bySetpos != null) {
      sb.append("; BYSETPOS=");
      sb.append(bySetpos.encode());
    }

    if (wkst >= 0) {
      sb.append("; WKST=");
      sb.append(weekday[wkst]);
    }

    ruleText = sb.toString();

    return ruleText;
  }

  /* ====================================================================
   *                        Convenience methods
   * ==================================================================== */

  /** Turn a week day of form "su", "mo" etc into an int
   *
   * @param val    String week day
   * @return int week day num or -1 if invalid
   */
  public static int toIntWday(String val) {
    for (int i = 0; i < weekday.length; i++) {
      if (weekday[i].equalsIgnoreCase(val)) {
        return i;
      }
    }

    return -1;
  }

  /* ====================================================================
   *                        Object methods
   * ==================================================================== */

  public String toString() {
    StringBuffer sb = new StringBuffer();

    sb.append("RecurRuleVO{id=");
    sb.append(id);
    sb.append(", rule=");
    sb.append(getRuleText());
    sb.append("}");

    return sb.toString();
  }

  /** The id does not take part in these comparisons because we're presumably
   * doing this to see if two objects are the same without doing a db retrieval.
   * That is, one of the objects in a comparison may not have the id set.
   *
   * <p>This is relevant for hashcode as the equals method requires that
   * hashcode return equal values for both objects in a comparison that are
   * to be considered equal.
   *
   * <p>As we have both objects to hand in the equals method we can check to
   * see if both ids are non-zero and use them in the comparison.
   */
  public int hashCode() {
    int hc = 1;

    if (id != 0) {
      hc *= id;
    }

    hc *= (freq + 1);

    if (until != null) {
      hc *= until.hashCode();
    }

    if (count > 0) {
      hc *= count;
    }

    if (interval > 0) {
      hc *= interval;
    }

    if (bySecond != null) {
      hc *= bySecond.hashCode();
    }

    if (byMinute != null) {
      hc *= byMinute.hashCode();
    }

    if (byHour != null) {
      hc *= byHour.hashCode();
    }

    if (byDay != null) {
      hc *= byDay.hashCode();
    }

    if (byMonthday != null) {
      hc *= byMonthday.hashCode();
    }

    if (byYearday != null) {
      hc *= byYearday.hashCode();
    }

    if (byWeekno != null) {
      hc *= byWeekno.hashCode();
    }

    if (byMonth != null) {
      hc *= byMonth.hashCode();
    }

    if (bySetpos != null) {
      hc *= bySetpos.hashCode();
    }

    if (wkst >= 0) {
      hc *= (wkst + 1);
    }

    return hc;
  }

  /**  Compares this object with the specified object for order.
   * Returns a negative integer, zero, or a positive integer as this object
   * is less than, equal to, or greater than the specified object.
   *
   */
  public int compareTo(Object o) {
    if (!(o instanceof RecurRuleVO)) {
      return -1;
    }

    RecurRuleVO that = (RecurRuleVO)o;

    if (that.id == id) {
      return 0;
    }

    if (that.id < id) {
      return 1;
    }

    return -1;
  }
}
