/* **********************************************************************
    Copyright 2003 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.rpi.cct.uwcal.synch.common;

import edu.rpi.cct.uwcal.calsvci.CalSvc;
import edu.rpi.cct.uwcal.calsvci.CalSvcI;
import edu.rpi.cct.uwcal.calsvci.CalSvcIPars;
import edu.rpi.cct.uwcal.common.URIgen;
import edu.rpi.cct.uwcal.icalendar.IcalTranslator;

import edu.washington.cac.calfacade.shared.CalFacadeException;
import edu.washington.cac.calfacade.shared.EventVO;
import edu.washington.cac.calfacade.shared.GuidVO;
import edu.washington.cac.calfacade.shared.SynchDataVO;
import edu.washington.cac.calfacade.shared.SynchInfoVO;
import edu.washington.cac.calfacade.shared.SynchStateVO;
import edu.washington.cac.calfacade.shared.UserAuth;

import net.fortuna.ical4j.model.component.VEvent;

import java.sql.Date;
import java.sql.Time;
import java.sql.Timestamp;
import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;
import org.apache.log4j.Logger;

/** Class to encapsulate the interactions between a synchronization client
 * and the uwcal suite. There are two sets of methods here, those that work
 * with Ical4j objects and those that work with UWCal objects.
 *
 * <p>There should be one call on init immediately after obtaining the
 * object.
 *
 * <p>Any further interactions must take place between an open and a close
 *
 * @author Mike Douglass    douglm@rpi.edu
 * @author Xavier Lawrence  jahia.org
 */
public class Synch {
  private boolean debug;

  private CalSvcI svci;

  /** The deviceId identifies the other end of the conversation. This is
   * something of an Syncml term
   */
  private String deviceId;

  private IcalTranslator trans;

  private SynchInfoVO synchInfo;

  private transient Logger log;

  /** Constructor
   *
   * @param account    String authenticated account name
   * @param deviceId   Identifies the client end.
   * @param uriGen     URIgen object used to provide ALTREP values.
   * @param debug      boolean true for some amount of debugging output.
   */
  public Synch(String account,
               String deviceId,
               URIgen uriGen, boolean debug) throws CalFacadeException {
    svci = new CalSvc();
    this.deviceId = deviceId;
    this.debug = debug;
    CalSvcIPars pars = new CalSvcIPars(account, UserAuth.noPrivileges,
                                       account,
                                       false,    // public
                                       deviceId, // synchId
                                       false, false, false, false,
                                       debug);
    svci.init(pars);

    /** If we were limiting our view to a particular filter we would set that
        now. */

    svci.open();

    try {
      trans = new IcalTranslator(svci, uriGen);

      synchInfo = svci.getSynchInfo();
      if (synchInfo == null) {
        synchInfo = new SynchInfoVO();
        synchInfo.setUser(svci.getUserVO());
        synchInfo.setDeviceId(deviceId);
        svci.addSynchInfo(synchInfo);
      }
    } finally {
      svci.close();
    }
  }

  /** Signal the start of a sequence of operations. These overlap transactions
   * in that there may be 0 to many transactions started and ended within an
   * open/close call and many open/close calls within a transaction.
   *
   * <p>The open close methods are mainly associated with web style requests
   * and open will usually be called early in the incoming request handling
   * and close will always be called on the way out allowing the interface an
   * opportunity to reacquire (on open) and release (on close) any resources
   * such as connections.
   */
  public void open() throws CalFacadeException {
    svci.open();
  }

  /** Call on the way out after handling a request..
   */
  public void close() throws CalFacadeException {
    svci.close();
  }

  /**
   * Starts a transaction on the backend persistent store of the calendar
   * server. This will make sure that all or none of the synchronization
   * updates succeed. If no transaction is required, simply avoid calling
   * this method.
   */
  public void beginTransaction() throws CalFacadeException {
    svci.beginTransaction();
  }

  /**
   * Commits the current transaction so all modifications to the backend
   * persistent store of the calendar server are made permanent.
   * (Should throw an Exception if no ongoing transaction exists)
   *
   * <p>We also update the synch state tables at this point.
   */
  public void commit() throws CalFacadeException {
    svci.updateSynchStates();

    if (synchInfo != null) {
      synchInfo.setLastsynch(new Timestamp(System.currentTimeMillis()));
      svci.updateSynchInfo(synchInfo);
    }

    svci.endTransaction();
  }

  /** Call if there has been an error during an update process.
   */
  public void rollbackTransaction() throws CalFacadeException {
    svci.rollbackTransaction();
  }

  /* --------------------------------------------------------------------
     Ical4j object methods
     -------------------------------------------------------------------- */

  /**
   * Returns all the Events associated with this conversation. The data
   * returned will be used for synchronization purposes and thus the state
   * of each event should also be returned.
   *
   * @return Collection of IcalSynchState objects.
   */
  public Collection getAllVevents() throws CalFacadeException {
    return makeVeventsCollection(getAllEvents(true));
  }

  /**
   * Returns all the changed Events associated with this conversation. The data
   * returned will be used for synchronization purposes and thus the state
   * of each event should also be returned.
   *
   * @return Collection of IcalSynchState objects.
   */
  public Collection getAllChangedVevents() throws CalFacadeException {
    return makeVeventsCollection(getAllEvents(false));
  }

  /**
   * Updates the event given as parameter using the provided new data.
   *
   * @param val          Updated vevent.
   * @param guid         Events guid
   * @return true        if the update was succesful
   */
  public boolean updateEvent(VEvent val, String guid) throws CalFacadeException {
    updateEvent(trans.toEvent(val, guid));

    return true;
  }

  /**
   * Deletes the event given as parameter..
   *
   * @param val          vevent defining event to delete.
   * @return true        if the update was succesful
   */
  public boolean deleteEvent(VEvent val, String guid) throws CalFacadeException {
    return deleteEvent(trans.toEvent(val, guid));
  }

  /* --------------------------------------------------------------------
     UWCal object methods
     -------------------------------------------------------------------- */

  /**
   * Returns all the Events associated with this conversation. The data
   * returned will be used for synchronization purposes and thus the state
   * of each event should also be returned.
   *
   * @param returnAll   boolean true if we return SYNCHRONIZED also.
   * @return Collection of SynchStateVO objects.
   */
  public Collection getAllEvents(boolean returnAll) throws CalFacadeException {
    Collection events = svci.getEvents();

    if (debug) {
      getLogger().debug("getAllEvents(): svci returned " + events.size());
    }

    Iterator it = events.iterator();
    Vector v = new Vector();

    while (it.hasNext()) {
      EventVO ev = (EventVO)it.next();

      SynchStateVO ss = svci.getSynchState(ev);
      if (debug) {
        //getLogger().debug("getAllEvents(): ev=" + ev);
        getLogger().debug("getAllEvents(): ss=" + ss);
      }

      if (ss == null) {
        /* This is the only place we add a synch state object. We do so
           because eiher we never synched or the event appeared since we last
           did a synch.
           */
        ss = new SynchStateVO();

        ss.setUser(svci.getUserVO());
        ss.setDeviceId(deviceId);
        ss.setEvent(ev);
        ss.setGuid(ev.getGuid());
        ss.setSynchState(SynchStateVO.NEW);

        svci.addSynchState(ss);
      } else if (ss.getSynchState() == SynchStateVO.CLIENT_DELETED) {
        // ignore this one
        ss = null;
      } else if (ss.getSynchState() == SynchStateVO.CLIENT_DELETED_UNDELIVERED) {
        /* We deliver a deleted state and will set the state to CLIENT_DELETED
         */
        ss.setSynchState(SynchStateVO.DELETED);
        ss.setEvent(null);
      } else if ((ss.getSynchState() == SynchStateVO.CLIENT_MODIFIED_UNDELIVERED) ||
                 (ss.getSynchState() == SynchStateVO.CLIENT_MODIFIED)) {
        if (!returnAll &&
             (ss.getSynchState() == SynchStateVO.CLIENT_MODIFIED)) {
          // ignore this one
          ss = null;
        } else {
          if (ss.getSynchState() == SynchStateVO.CLIENT_MODIFIED) {
            ss.setSynchState(SynchStateVO.SYNCHRONIZED);
          } else {
            /* We deliver a modified state and will set the state to CLIENT_MODIFIED
             */
            ss.setSynchState(SynchStateVO.MODIFIED);
          }
          svci.getSynchData(ss);

          SynchDataVO synchData = ss.getSynchData();

          if (synchData == null) {
            if (debug) {
              getLogger().debug("No event data present for id " + ev.getId());
            }
            throw new CalFacadeException("No event data present for id " + ev.getId());
          }

          String evData = synchData.getEventData();
          Collection c = trans.toEvent(evData);
          if (c.size() != 1) {
            throw new CalFacadeException("Invalid event data: " + evData);
          }

          Iterator eit = c.iterator();

          ss.setEvent((EventVO)eit.next());
        }
      } else {
        /** If the event state is SYNCHRONIZED we look to see if the event
         * is modified by checking the  lastsynch against the event lastmod.
         */
        if (ss.getSynchState() == SynchStateVO.SYNCHRONIZED) {
          /** See if the event has been modified since last synch */

          Timestamp lastsynch = synchInfo.getLastsynch();
          if ((lastsynch == null) ||
               (lastsynch.compareTo(ev.getLastmod()) <= 0)) {
            if (debug) {
              getLogger().debug("Event modified for id " + ev.getId());
            }

            ss.setSynchState(SynchStateVO.MODIFIED);
          }
        }

        if (!returnAll &&
            (ss.getSynchState() == SynchStateVO.SYNCHRONIZED)) {
          // Don't return this one.
          ss = null;
        }
      }

      if (ss != null) {
        v.add(ss);
      }
    }

    /** Append any deleted events
     */
    v.addAll(svci.getDeletedSynchStates());

    return v;
  }

  /**
   * Updates the event given as parameter using the provided new data.
   *
   * <p>This method will be called after getNewGuid for a new event.
   *
   * @param val          Updated event.
   * @return true        if the update was succesful
   */
  public boolean updateEvent(EventVO val) throws CalFacadeException {
    /** Get the synch state - will add it if it doesn't exist at the moment.
     */
    SynchStateVO ss = getSynchState(val);

    /* We should check the access here - for the moment just see if it's
       public
       */
    if (!val.isPublic()) {
      svci.updateEvent(val, null);
    } else {
      /* Convert to an icalendar string and save it */
      SynchDataVO sd = new SynchDataVO(ss.getUser(), val.getId(),
                                       trans.toStringIcal(val));

      ss.setSynchData(sd);
      svci.setSynchData(ss);
    }

    return true;
  }

  /** Creates a new event based on the given data.
   *
   * @param data The Object containing event's data
   * @return True if the creation was successful
   */
  public boolean addEvent(EventVO val) throws CalFacadeException {
    svci.addEvent(val);

    return true;
  }

  /**
   * Deletes a particular event.
   *
   * @param val     The event to delete
   * @return boolean true if the deletion was successful
   */
  public boolean deleteEvent(EventVO val) throws CalFacadeException {
    EventVO ev = svci.getEvent(val.getId());

    if (ev == null) {
      return false;
    }

    if (ev.getKind() == EventVO.kindEntry) {
      CalSvcI.DelEventResult der = svci.deleteEvent(ev, true);
      return der.eventDeleted;
    }

    if (ev.getKind() == EventVO.kindAdded) {
      svci.deleteEventRef(ev);
    } else if (ev.getKind() == EventVO.kindUndeletable) {
      svci.deleteSubscribedEvent(ev);
    }

    return true;
  }

  /**
   * Find an event given the guid.
   *
   * @param val     The guid - must be non-null
   * @return EventVO if found or null
   */
  public EventVO findEvent(String val) throws CalFacadeException {
    GuidVO guid = GuidVO.valueOf(val);

    if (!svci.getSysid().equals(guid.getSysid())) {
      throw new CalFacadeException("Invalid uwcal guid: " + val);
    }

    return svci.getEvent(guid.getId());
  }

  /** Obtain a new guid for an event.
   *
   * <p>The synch process requires a guid for an event which is to be
   * created in advance of the actual data.
   *
   * <p>We will create a dummy event with a 0 start and end date and time.
   *
   * <p>This process only works if we (nearly) always update the newly
   * created event.
   *
   * <p>The event should be flagged as temporary and have its status changed
   * when the data arrives.
   *
   * @param val     The guid
   * @return EventVO if found or null
   */
  public String getNewGuid() throws CalFacadeException {
    EventVO ev = new EventVO();

    ev.setStartdate(new Date(0));
    ev.setStarttime(new Time(0));
    ev.setEnddate(new Date(0));
    ev.setEndtime(new Time(0));
    ev.setShortdesc("Temp");

    svci.addEvent(ev);

    return ev.getGuid();
  }

  /* --------------------------------------------------------------------
     Convenience methods
     -------------------------------------------------------------------- */

  /** Return the current IcalTranslator
   */
  public IcalTranslator getTranslator() {
    return trans;
  }

  /* ====================================================================
                       Private methods
     ==================================================================== */

  /**
   * Make a collction of IcalSynchState from a Collection of SynchStateVO
   *
   * @param events    Collection of SynchStateVO
   * @return Collection of IcalSynchState objects.
   */
  private Collection makeVeventsCollection(Collection events)
          throws CalFacadeException {
    Iterator it = events.iterator();
    Vector v = new Vector();

    while (it.hasNext()) {
      SynchStateVO ss = (SynchStateVO)it.next();

      IcalSynchState iss = new IcalSynchState();

      iss.setEvent(ss.getEvent());
      iss.setGuid(ss.getGuid());
      iss.setSynchState(ss.getSynchState());
      iss.setUser(ss.getUser());
      iss.setVevent(trans.toIcalEvent(ss.getEvent()));

      v.add(iss);
    }

    return v;
  }

  /** Get a synch state object forthe given event. If one does not exist
   * create it now and set it's state to NEW.
   *
   * <p>This is the only place we add a synch state object. We do so
   * because eiher we never synched or the event appeared since we last
   * did a synch
   *
   * @param ev            EventVO object .
   * @return SynchStateVO object
   */
  private SynchStateVO getSynchState(EventVO ev) throws CalFacadeException {
    SynchStateVO ss = svci.getSynchState(ev);

    if (ss != null) {
      return ss;
    }

    ss = new SynchStateVO();

    ss.setUser(svci.getUserVO());
    ss.setDeviceId(deviceId);
    ss.setEvent(ev);
    ss.setGuid(ev.getGuid());
    ss.setSynchState(SynchStateVO.NEW);

    svci.addSynchState(ss);

    return ss;
  }

  /** Get a logger for messages
   */
  private Logger getLogger() {
    if (log == null) {
      log = Logger.getLogger(this.getClass());
    }

    return log;
  }
}


