/* **********************************************************************
    Copyright 2003 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.tests.rpi.cct.uwcal.calsvci;

import edu.rpi.cct.uwcal.calsvci.CalSvc;
import edu.rpi.cct.uwcal.calsvci.CalSvcI;
import edu.rpi.cct.uwcal.calsvci.CalSvcIPars;

import edu.washington.cac.calfacade.shared.CalFacadeDefs;
import edu.washington.cac.calfacade.shared.EventVO;
import edu.washington.cac.calfacade.shared.KeywordVO;
import edu.washington.cac.calfacade.shared.LocationVO;
import edu.washington.cac.calfacade.shared.SponsorVO;
import edu.washington.cac.calfacade.shared.UserVO;

import java.sql.Date;
import java.sql.Time;
import java.sql.Timestamp;
import java.text.SimpleDateFormat;
import java.util.Locale;
import junit.framework.TestCase;

/** Conveneience classes for test suite
 *
 * @author Mike Douglass       douglm@rpi.edu
 */
public class CalSvcTestUtil extends TestCase {
  /* public events */
  private CalSvcI pubSvci;

  /* private events */
  private CalSvcI privSvci;

  private boolean debug = true;

  public CalSvcTestUtil(boolean debug) {
    this.debug = debug;
  }

  /* ====================================================================
   *                       Misc methods.
   * ==================================================================== */

  public CalSvcI getSvci(String user,
                        int access,
                        boolean publicEvents) {
    CalSvcI svci = null;

    try {
      if (publicEvents) {
        svci = pubSvci;
      } else {
        svci = privSvci;
      }

      if (svci == null) {
        svci = new CalSvc();
        CalSvcIPars pars = new CalSvcIPars(user, access, user, publicEvents,
                                           null, // synch
                                           false, false, false, false,
                                           debug);
        svci.init(pars);
      }

      if (publicEvents) {
        pubSvci = svci;
      } else {
        privSvci = svci;
      }
    } catch (Throwable t) {
      t.printStackTrace();
      fail(t.getMessage());
    }

    return svci;
  }

  public CalSvcI getSvci(boolean publicEvents) {
    if (publicEvents) {
      return pubSvci;
    }

    return privSvci;
  }

  public void open(boolean publicEvents) {
    try {
      CalSvcI svci = getSvci(publicEvents);

      svci.open();
      svci.beginTransaction();
    } catch (Throwable t) {
      t.printStackTrace();
      fail(t.getMessage());
    }
  }

  public void close(boolean publicEvents) {
    try {
      CalSvcI svci = getSvci(publicEvents);

      svci.endTransaction();
      svci.close();
    } catch (Throwable t) {
      t.printStackTrace();
      fail(t.getMessage());
    }
  }

  /* ====================================================================
   *                       location methods.
   * ==================================================================== */

  /** Add a single location for later use
   *
   * @return int  key of first added (remainder should follow)
   */
  public int addLocation(String address,
                         String subaddress,
                         String link,
                         String creator,
                         boolean isPublic) {
    LocationVO loc = new LocationVO(CalFacadeDefs.unsavedItemKey,
                                    address,
                                    subaddress,
                                    link,
                                    new UserVO(creator),
                                    isPublic);
    try {
      getSvci(isPublic).addLocation(loc);
      log("Created location, id=" + loc.getId());
      return loc.getId();
    } catch (Throwable t) {
      t.printStackTrace();
      fail("Exception adding location: " + loc);
      return CalFacadeDefs.unsavedItemKey;
    }
  }

  /** Get a location which must exist
   *
   * @param id           int key of object
   * @return LocationVO  or null for failure
   */
  public LocationVO getLocation(int id, boolean publicEntity) {
    return getLocation(id, true, publicEntity);
  }

  /** Get a location which may or may not exist
   *
   * @param id           int key of object
   * @param mustExist    boolean true if object is required to exist
   * @return LocationVO  or null for failure
   */
  public LocationVO getLocation(int id, boolean mustExist, boolean publicEntity) {
    try {
      LocationVO obj = getSvci(publicEntity).getLocation(id);
      if (mustExist && (obj == null)) {
        fail("Location " + id + " does not exist");
      }

      return obj;
    } catch (Throwable t) {
      t.printStackTrace();
      fail("Exception getting location: " + id);
      return null;
    }
  }

  /* ====================================================================
   *                       sponsor methods.
   * ==================================================================== */

  /** Add a single sponsor for later use
   *
   * @return int  key of first added (remainder should follow)
   */
  public int addSponsor(String name,
                        String phone,
                        String email,
                        String link,
                        String creator,
                        boolean isPublic) {
    SponsorVO sp = new SponsorVO(CalFacadeDefs.unsavedItemKey,
                                 name,
                                 phone,
                                 email,
                                 link,
                                 new UserVO(creator),
                                 isPublic);

    try {
      getSvci(isPublic).addSponsor(sp);
      log("Created sponsor, id=" + sp.getId());
      return sp.getId();
    } catch (Throwable t) {
      t.printStackTrace();
      fail("Exception adding sponsor: " + sp);
      return CalFacadeDefs.unsavedItemKey;
    }
  }

  /** Get a sponsor which must exist
   *
   * @param id           int key of object
   * @return LocationVO  or null for failure
   */
  public SponsorVO getSponsor(int id, boolean publicEntity) {
    return getSponsor(id, true, publicEntity);
  }

  /** Get a sponsor which may or may not exist
   *
   * @param id           int key of object
   * @param mustExist    boolean true if object is required to exist
   * @return LocationVO  or null for failure
   */
  public SponsorVO getSponsor(int id, boolean mustExist, boolean publicEntity) {
    try {
      SponsorVO obj = getSvci(publicEntity).getPublicSponsor(id);
      if (mustExist && (obj == null)) {
        fail("Sponsor " + id + " does not exist");
      }

      return obj;
    } catch (Throwable t) {
      t.printStackTrace();
      fail("Exception getting sponsor: " + id);
      return null;
    }
  }

  /* ====================================================================
   *                       keyword methods.
   * ==================================================================== */

  /** Add a single keyword for later use
   *
   * @return int  key of first added (remainder should follow)
   */
  public int addKeyword(String word,
                        String description,
                        String creator,
                        boolean isPublic) {
    KeywordVO key = new KeywordVO(CalFacadeDefs.unsavedItemKey,
                                  word,
                                  null,
                                  description,
                                  isPublic,
                                  new UserVO(creator));

    try {
      getSvci(isPublic).addKeyword(key);
      log("Created keyword, id=" + key.getId());
      return key.getId();
    } catch (Throwable t) {
      t.printStackTrace();
      fail("Exception adding keyword: " + key);
      return CalFacadeDefs.unsavedItemKey;
    }
  }

  /** Add a single keyword with id for later use
   *
   */
  public int addKeyword(int id, String word,
                        String description,
                        String creator,
                        boolean isPublic) {
    KeywordVO key = new KeywordVO(id,
                                  word,
                                  null,
                                  description,
                                  isPublic,
                                  new UserVO(creator));

    try {
      getSvci(isPublic).addKeywordWithKey(key);
      log("Created keyword, id=" + key.getId());
      return key.getId();
    } catch (Throwable t) {
      t.printStackTrace();
      fail("Exception adding keyword: " + key);
      return CalFacadeDefs.unsavedItemKey;
    }
  }

  /** Get a keyword which must exist
   *
   * @param id           int key of object
   * @return KeywordVO  or null for failure
   */
  public KeywordVO getKeyword(int id, boolean publicEntity) {
    return getKeyword(id, true, publicEntity);
  }

  /** Find a keyword which must exist
   *
   * @param id           int key of object
   * @return KeywordVO    or null for failure
   */
  public KeywordVO getKeyword(String wd, boolean publicEntity) {
    try {
      KeywordVO par = new KeywordVO();
      par.setWord(wd);
      KeywordVO obj = getSvci(publicEntity).findKeyword(par);
      if (obj == null) {
        fail("Keyword " + wd + " does not exist");
      }

      return obj;
    } catch (Throwable t) {
      t.printStackTrace();
      fail("Exception getting keyword: " + wd);
      return null;
    }
  }

  /** Get a keyword which may or may not exist
   *
   * @param id           int key of object
   * @param mustExist    boolean true if object is required to exist
   * @return LocationVO  or null for failure
   */
  public KeywordVO getKeyword(int id, boolean mustExist, boolean publicEntity) {
    try {
      KeywordVO obj = getSvci(publicEntity).getPublicKeyword(id);
      if (mustExist && (obj == null)) {
        fail("Keyword " + id + " does not exist");
      }

      return obj;
    } catch (Throwable t) {
      t.printStackTrace();
      fail("Exception getting keyword: " + id);
      return null;
    }
  }

  /* ====================================================================
   *                       event methods.
   * ==================================================================== */

  /** Add a single event for later use. Dates should be of form
   * yyyy-MM-dd HH:mm:ss
   *
   * @return int  key of new entity
   */
  public int addEvent(String sdesc,
                      String ldesc,
                      String sdatetime,
                      String edatetime,
                      LocationVO loc,
                      SponsorVO sp,
                      KeywordVO key,
                      String link,
                      String creator,
                      boolean isPublic) throws Throwable {
    EventVO ev = new EventVO(CalFacadeDefs.unsavedItemKey,
                             sdesc,
                             ldesc,
                             null,
                             null,
                             null,
                             null,
                             isPublic,
                             link,
                             new UserVO(creator),
                             "free",
                             null,
                             loc,
                             sp,
                             1,
                             new Timestamp(System.currentTimeMillis()),
                             new Timestamp(System.currentTimeMillis()));

    SimpleDateFormat fmt = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss",
                                                Locale.US);
    long millis = fmt.parse(sdatetime).getTime();

    ev.setStartdate(new Date(millis));
    ev.setStarttime(new Time(millis));

    millis = fmt.parse(edatetime).getTime();

    ev.setEnddate(new Date(millis));
    ev.setEndtime(new Time(millis));

    if (key != null) {
      ev.addKeyword(key);
    }

    try {
      getSvci(isPublic).addEvent(ev);
      log("Created event, id=" + ev.getId());
      return ev.getId();
    } catch (Throwable t) {
      t.printStackTrace();
      fail("Exception adding evnt: " + ev);
      return CalFacadeDefs.unsavedItemKey;
    }
  }

  /** Get an event which may or may not exist
   *
   * @param id           int key of object
   * @param mustExist    boolean true if object is required to exist
   * @return LocationVO  or null for failure
   */
  public EventVO getEvent(int id, boolean mustExist, boolean publicEntity) {
    try {
      EventVO obj = getSvci(publicEntity).getEvent(id);
      if (mustExist && (obj == null)) {
        fail("Event " + id + " does not exist");
      }

      return obj;
    } catch (Throwable t) {
      t.printStackTrace();
      fail("Exception getting Event: " + id);
      return null;
    }
  }

  private void log(String msg) {
    System.out.println(this.getClass().getName() + ": " + msg);
  }
}

