/* **********************************************************************
    Copyright 2003 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.tests.rpi.cct.uwcal.icalendar;

import edu.tests.rpi.cct.uwcal.calsvci.CalSvcTestUtil;

import edu.rpi.cct.uwcal.icalendar.IcalTranslator;
import edu.rpi.cct.uwcal.common.URIgen;
import edu.rpi.cct.uwcal.webcommon.UWCalWebURIgen;
import edu.rpi.sss.util.log.Log4jOutputStream;

import edu.washington.cac.calfacade.shared.CalFacadeDefs;
import edu.washington.cac.calfacade.shared.EventVO;
import edu.washington.cac.calfacade.shared.KeywordVO;
import edu.washington.cac.calfacade.shared.LocationVO;
import edu.washington.cac.calfacade.shared.SponsorVO;
import edu.washington.cac.calfacade.shared.UserAuth;
import edu.washington.cac.calfacade.shared.UserVO;

import java.sql.Date;
import java.sql.Time;
import java.sql.Timestamp;
import java.util.Collection;
import java.util.Iterator;

import junit.framework.TestCase;

import net.fortuna.ical4j.data.CalendarOutputter;
import net.fortuna.ical4j.model.Calendar;

/** Test the IcalTranslator class
 *
 * <p>We'll manufacture an EventVO object, feed it through the translator
 * then send it back again. The end result should be equal.
 *
 * @author Mike Douglass       douglm@rpi.edu
 */
public class IcalTranslatorTest extends TestCase {
  private CalSvcTestUtil svciUtil;

  private boolean debug = true;

  private static String user = "caladmin";

  private static final String[] icalText = {
    "BEGIN:VCALENDAR",
    "PRODID:UW calendar V2.0",
    "VERSION:2.0",
    "BEGIN:VEVENT",
    "CATEGORIES:",
    "CLASS:PRIVATE",
    "CREATED;VALUE=DATE-TIME:20041210T160137Z",
    "DESCRIPTION:Dummy ical event description",
    "DTEND;VALUE=DATE-TIME:20041118T170000",
    "DTSTART;VALUE=DATE-TIME:20041118T160000",
    "LAST-MODIFIED;VALUE=DATE-TIME:20041210T160137Z",
    "LOCATION;ALTREP=\"http://cal.rpi.edu/location.do?locationId=14\":Test user s",
    " ynch location address\\, Test user synch location subaddress\\, Url: http:/",
    " /dummy.synch.link.edu",
    "ORGANIZER:caladmin",
    "SEQUENCE:1",
    "SUMMARY:Dummy ical event  summary",
    "UID:12-calendar@mysite.edu",
    "URL:http://www.rpi.edu",
    "END:VEVENT",
    "END:VCALENDAR",
  };

  public void testIcalTranslator() {
    try {
      svciUtil = new CalSvcTestUtil(debug);

      /* Init public svci */
      svciUtil.getSvci(user, UserAuth.superUser, true);

      /* Init personal svci */
      svciUtil.getSvci(user, UserAuth.superUser, false);

      /* Add a private location */
      int locKey = svciUtil.addLocation("Test user synch location address",
                                        "Test user synch location subaddress",
                                        "http://dummy.synch.link.edu",
                                        user,
                                        false);

      if (locKey < 0) {
        fail("previous step failed");
      }

      /* Add a private sponsor
      int spKey = svciUtil.addSponsor("Test user synch sponsor name",
                                      "Test user synch sponsor phone",
                                      "Test user synch sponsor email",
                                      "http://dummy.synch.link.edu",
                                      user,
                                      false);

      if (spKey < 0) {
        fail("previous step failed");
      }*/

      /* Add a public category (keyword) */
      int kKey = svciUtil.addKeyword("Test user synch keyword",
                                      "Test user synch keyword description",
                                      user,
                                      true);

      if (kKey < 0) {
        fail("previous step failed");
      }

      LocationVO loc = svciUtil.getLocation(locKey, false);
      KeywordVO key = svciUtil.getKeyword(kKey, true);

      int eventId = svciUtil.addEvent("Dummy ical event  summary",
                                     "Dummy ical event description",
                                     "2004-11-18 16:00:00",
                                     "2004-11-18 17:00:00",
                                     loc,
                                     null,
                                     key,
                                     "http://www.rpi.edu",
                                     user,
                                     false);

      EventVO ev = svciUtil.getEvent(eventId, true, false);

      URIgen urigen = new UWCalWebURIgen("http://cal.rpi.edu");

      IcalTranslator icalTrans = new IcalTranslator(svciUtil.getSvci(false),
                                                    urigen, debug);

      Calendar ical = icalTrans.toIcal(ev);

      CalendarOutputter calOut = new CalendarOutputter(true);

      Log4jOutputStream out = new Log4jOutputStream();

      System.out.println("===========================================");
      calOut.output(ical, out);
      System.out.println("===========================================");

      Collection c = makeEvents(icalTrans, icalText);
      Iterator it = c.iterator();

      while (it.hasNext()) {
        EventVO e = (EventVO)it.next();

        log(e.toString());
      }
    } catch (Throwable t) {
      t.printStackTrace();
      fail("Exception getting service interface: " + t.getMessage());
    }
  }

  /* ====================================================================
   *                       Private methods.
   * ==================================================================== */

  private Collection makeEvents(IcalTranslator icalTrans,
                                String[] calText) throws Throwable {
    StringBuffer sb = new StringBuffer();

    for (int i = 0; i < calText.length; i++) {
      sb.append(calText[i]);
      sb.append("\r\n");
    }

    return icalTrans.toEvent(sb.toString());
  }

  private void log(String msg) {
    System.out.println(this.getClass().getName() + ": " + msg);
  }
}

