package edu.rpi.cct.uwcal.tools.dump;

import edu.washington.cac.calfacade.shared.AdminGroupVO;
import edu.washington.cac.calfacade.shared.AuthUserVO;
import edu.washington.cac.calfacade.shared.AuthUserPrefsVO;
import edu.washington.cac.calfacade.shared.EventVO;
import edu.washington.cac.calfacade.shared.EventRefVO;
import edu.washington.cac.calfacade.shared.filter.AndFilterVO;
import edu.washington.cac.calfacade.shared.filter.CreatorFilterVO;
import edu.washington.cac.calfacade.shared.filter.FilterVO;
import edu.washington.cac.calfacade.shared.filter.KeywordFilterVO;
import edu.washington.cac.calfacade.shared.filter.LocationFilterVO;
import edu.washington.cac.calfacade.shared.filter.NotFilterVO;
import edu.washington.cac.calfacade.shared.filter.OrFilterVO;
import edu.washington.cac.calfacade.shared.filter.SponsorFilterVO;
import edu.washington.cac.calfacade.shared.KeywordAttrVO;
import edu.washington.cac.calfacade.shared.KeywordVO;
import edu.washington.cac.calfacade.shared.LocationVO;
import edu.washington.cac.calfacade.shared.SponsorVO;
import edu.washington.cac.calfacade.shared.UserVO;

import java.io.FileWriter;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.sql.Date;
import java.sql.Time;
import java.sql.Timestamp;
import java.util.Collection;
import java.util.Iterator;
import java.util.Set;

/** Class which emits a generic sql file.
 *
 * @author Mike Douglass   douglm@rpi.edu
 * @version 1.0
 */
public class SqlOut implements RestoreIntf {
  /** The sql file
   */
  private String sqlFileName;
  private Writer sqlOut;

  /** This needs the CalEnv class to be properly set up.
   */
  private int maxDbStringLength = 0;

  public SqlOut(String sqlFileName) throws Throwable {
    this.sqlFileName = sqlFileName;
    sqlOut = new FileWriter(sqlFileName);
  }

  /** Call after any init phase
   */
  public void open() throws Throwable {
  }

  /** Call at end of restoring to finish up. Will restore any cached values.
   */
  public void close() throws Throwable {
    if (sqlOut != null) {
      sqlOut.close();
    }
  }

  /** Restore users NOT subscriptions
   */
  public void restoreUser(UserVO o) throws Throwable {
    sqlOut.write("INSERT INTO users VALUES (");
    sqlInt(o.getId());
    sqlOut.write(',');
    sqlString(o.getAccount());
    sqlOut.write(',');
    sqlTimestamp(o.getCreated());
    sqlOut.write(',');
    sqlTimestamp(o.getLogon());
    sqlOut.write(',');
    sqlTimestamp(o.getLastAccess());
    sqlOut.write(',');
    sqlTimestamp(o.getLastModify());
    sqlOut.write(");");

    sqlOut.write('\n');
  }

  /** Restore an admin group and its members - though not the user entries nor
   * the authuser entries.
   */
  public void restoreAdminGroup(AdminGroupVO o) throws Throwable {
    sqlOut.write("INSERT INTO adminGroups VALUES (");
    sqlString(o.getName());
    sqlOut.write(',');
    sqlString(o.getDescription());
    sqlOut.write(',');
    sqlAccount(o.getGroupOwner());
    sqlOut.write(',');
    sqlAccount(o.getOwner());
    sqlOut.write(");");

    sqlOut.write('\n');

    Collection mbrs = o.getGroupMembers();

    if (mbrs != null) {
      Iterator uit = mbrs.iterator();

      while (uit.hasNext()) {
        UserVO u = (UserVO)uit.next();

        sqlOut.write("INSERT INTO adminGroupMembers VALUES (");
        sqlString(o.getName());
        sqlOut.write(',');
        sqlAccount(u);
        sqlOut.write(");");

        sqlOut.write('\n');
      }
    }
  }

  /** Restore an auth user and preferences
   */
  public void restoreAuthUser(AuthUserVO o) throws Throwable {
    sqlOut.write("INSERT INTO auth VALUES (");
    sqlInt(o.getUser().getId());
    sqlOut.write(',');
    sqlInt(o.getUsertype());
    sqlOut.write(',');
    sqlString(o.getLastname());
    sqlOut.write(',');
    sqlString(o.getFirstname());
    sqlOut.write(',');
    sqlString(o.getPhone());
    sqlOut.write(',');
    sqlString(o.getEmail());
    sqlOut.write(',');
    sqlString(o.getDept());
    sqlOut.write(");");

    sqlOut.write('\n');
  }

  /** Restore an event and event keyword entries
   */
  public void restoreEvent(EventVO o) throws Throwable {
    sqlOut.write("INSERT INTO events VALUES (");
    sqlInt(o.getId());
    sqlOut.write(',');
    sqlTimestamp(o.getLastmod());
    sqlOut.write(',');
    sqlInt(o.getSeq());
    sqlOut.write(',');
    sqlBool(o.getPublic());
    sqlOut.write(',');
    sqlTimestamp(o.getCreated());
    sqlOut.write(',');
    sqlDate(o.getStartdate());
    sqlOut.write(',');
    sqlTime(o.getStarttime());
    sqlOut.write(',');
    sqlDate(o.getEnddate());
    sqlOut.write(',');
    sqlTime(o.getEndtime());
    sqlOut.write(',');
    sqlString(o.getShortdesc());
    sqlOut.write(',');


    String ld = o.getLongdesc();
    String ld1 = null;

    if (maxDbStringLength > 0) {
      if ((ld != null) && (ld.length() > maxDbStringLength)) {
        ld1 = ld.substring(maxDbStringLength);
        ld = ld.substring(0, maxDbStringLength);
      }
    }

    sqlString(ld);
    sqlOut.write(',');
    sqlString(ld1);
    sqlOut.write(',');
    sqlString(null);  //longdesc2
    sqlOut.write(',');

    sqlString(o.getLink());
    sqlOut.write(',');
    sqlString("F");
    sqlOut.write(',');
    sqlInt(o.getLocation().getId());
    sqlOut.write(',');
    sqlInt(o.getSponsor().getId());
    sqlOut.write(',');
    sqlAccount(o.getCreator());
    sqlOut.write(',');
//      sqlChar(o.getRecurringStatus());
//      sqlOut.write(',');
    sqlString(o.getCost());
    sqlOut.write(");");

    sqlOut.write('\n');

    restoreEventKeys(o);
  }

  /** Restore event ref
   */
  public void restoreEventRef(EventRefVO o) throws Throwable {
    sqlOut.write("INSERT INTO eventrefs VALUES (");
    sqlInt(o.getEventid());
    sqlOut.write(',');
    sqlUserId(o.getUser());
    sqlOut.write(',');
    sqlChar(o.getPurpose());
    sqlOut.write(");");

    sqlOut.write('\n');
  }

  /** Restore keyword and attributes
   */
  public void restoreKeyword(KeywordVO o) throws Throwable {
    sqlOut.write("INSERT INTO keywords VALUES (");
    sqlInt(o.getId());
    sqlOut.write(',');
    sqlString(o.getWord());
    sqlOut.write(',');
    sqlString(o.getTitle());
    sqlOut.write(',');
    sqlString(o.getDescription());
    sqlOut.write(',');
    sqlAccount(o.getCreator());
    sqlOut.write(',');
    sqlBool(o.isPublic());
    sqlOut.write(");");

    sqlOut.write('\n');

    Set s = o.getAttributes();

    if ((s != null) && (s.size() > 0)) {
      Iterator it = s.iterator();

      while (it.hasNext()) {
        KeywordAttrVO kao = (KeywordAttrVO)it.next();

        sqlOut.write("INSERT INTO keyword_attrs VALUES (");
        sqlInt(kao.getId());
        sqlOut.write(',');
        sqlInt(kao.getKeyword().getId());
        sqlOut.write(',');
        sqlString(kao.getName());
        sqlOut.write(',');
        sqlString(kao.getVal());
        sqlOut.write(");");

        sqlOut.write('\n');
      }
    }
  }

  /** Restore location
   */
  public void restoreLocation(LocationVO o) throws Throwable {
    sqlOut.write("INSERT INTO locations VALUES (");
    sqlInt(o.getId());
    sqlOut.write(',');
    sqlString(o.getAddress());
    sqlOut.write(',');
    sqlString(o.getSubaddress());
    sqlOut.write(',');
    sqlString(o.getLink());
    sqlOut.write(',');
    sqlAccount(o.getCreator());
    sqlOut.write(',');
    sqlBool(o.isPublic());
    sqlOut.write(");");

    sqlOut.write('\n');
  }

  /** Restore sponsor
   */
  public void restoreSponsor(SponsorVO o) throws Throwable {
    sqlOut.write("INSERT INTO sponsors VALUES (");
    sqlInt(o.getId());
    sqlOut.write(',');
    sqlString(o.getName());
    sqlOut.write(',');
    sqlString(o.getPhone());
    sqlOut.write(',');
    sqlString(o.getEmail());
    sqlOut.write(',');
    sqlString(o.getLink());
    sqlOut.write(',');
    sqlAccount(o.getCreator());
    sqlOut.write(',');
    sqlBool(o.isPublic());
    sqlOut.write(");");

    sqlOut.write('\n');
  }

  /** Restore filter
   */
  public void restoreFilter(FilterVO o) throws Throwable {
    /* Set this according to class of filter.
     */
    char typeFlag;
    int ref = 0;

    if (o instanceof AndFilterVO) {
      typeFlag = 'A';
    } else if (o instanceof KeywordFilterVO) {
      typeFlag = 'K';

      KeywordVO k = ((KeywordFilterVO)o).getKeyword();

      ref = k.getId();
    } else if (o instanceof CreatorFilterVO) {
      typeFlag = 'C';

      UserVO u = ((CreatorFilterVO)o).getCreator();

      ref = u.getId();
    } else if (o instanceof LocationFilterVO) {
      typeFlag = 'L';

      LocationVO l = ((LocationFilterVO)o).getLocation();

      ref = l.getId();
    } else if (o instanceof OrFilterVO) {
      typeFlag = 'O';
    } else if (o instanceof SponsorFilterVO) {
      typeFlag = 'S';

      SponsorVO sp = ((SponsorFilterVO)o).getSponsor();

      ref = sp.getId();
    } else if (o instanceof NotFilterVO) {
      typeFlag = 'N';

      FilterVO f = ((NotFilterVO)o).getFilter();

      ref = f.getId();
    } else {
      throw new Exception("Unimplemented filter " + o);
    }

    sqlOut.write("INSERT INTO calendars VALUES (");
    sqlInt(o.getId());
    sqlOut.write(',');
    sqlChar(typeFlag);
    sqlOut.write(',');
    sqlString(o.getName());
    sqlOut.write(',');
    sqlString(o.getTitle());
    sqlOut.write(',');
    sqlString(o.getDescription());
    sqlOut.write(',');
    sqlInt(ref);
    sqlOut.write(',');
    sqlInt(o.getParent().getId());
    sqlOut.write(");");

    sqlOut.write('\n');
  }

  /** Subscribe user to filter.
   */
  public void subscribe(UserVO user, FilterVO f) throws Throwable {
    sqlOut.write("INSERT INTO subscriptions VALUES (");
    sqlUserId(user);
    sqlOut.write(',');
    sqlInt(f.getId());
    sqlOut.write(");");

    sqlOut.write('\n');
  }

  /** Restore db lastmods
   */
  public void restoreDbLastmod(DumpDefs.DbLastmodVO o) throws Throwable {
    sqlOut.write("INSERT INTO lastmods VALUES (");
    sqlString(o.getDbname());
    sqlOut.write(',');
    sqlTimestamp(o.getLastmod());
    sqlOut.write(");");

    sqlOut.write('\n');
  }

  /** ===================================================================
   *                       Private methods
   *  =================================================================== */

  /** Iterate over the preferences multiple times to build each set of
   * inserts separately.
   */
  private void restoreAuthPrefs(AuthUserPrefsVO o) throws Exception {
    /* first the prefs */

    sqlOut.write("INSERT INTO authprefs VALUES (");
    sqlInt(o.getAuthUser().getUser().getId());
    sqlOut.write(',');
    sqlBool(o.getAutoAddKeywords());
    sqlOut.write(',');
    sqlBool(o.getAutoAddLocations());
    sqlOut.write(',');
    sqlBool(o.getAutoAddSponsors());
    sqlOut.write(");");

    sqlOut.write('\n');

    /* preferred keywords */
    Set kpref = o.getPreferredKeywords();

    if (kpref != null) {
      Iterator it = kpref.iterator();

      while (it.hasNext()) {
        KeywordVO p = (KeywordVO)it.next();

        sqlOut.write("INSERT INTO authprefKeywords VALUES (");
        sqlInt(o.getAuthUser().getUser().getId());
        sqlOut.write(',');
        sqlInt(p.getId());
        sqlOut.write(");");

        sqlOut.write('\n');
      }
    }

    /* preferred locations */
    Set lpref = o.getPreferredLocations();

    if (lpref != null) {
      Iterator it = lpref.iterator();

      while (it.hasNext()) {
        LocationVO p = (LocationVO)it.next();

        sqlOut.write("INSERT INTO authprefLocations VALUES (");
        sqlInt(o.getAuthUser().getUser().getId());
        sqlOut.write(',');
        sqlInt(p.getId());
        sqlOut.write(");");

        sqlOut.write('\n');
      }
    }

    /* preferred sponsors */
    Set spref = o.getPreferredSponsors();

    if (spref != null) {
      Iterator it = spref.iterator();

      while (it.hasNext()) {
        SponsorVO p = (SponsorVO)it.next();

        sqlOut.write("INSERT INTO authprefSponsors VALUES (");
        sqlInt(o.getAuthUser().getUser().getId());
        sqlOut.write(',');
        sqlInt(p.getId());
        sqlOut.write(");");

        sqlOut.write('\n');
      }
    }
  }

  private void restoreEventKeys(EventVO ev) throws Exception {
    Collection evks = ev.getKeywords();
    int evid = ev.getId();

    Iterator it = evks.iterator();

    while (it.hasNext()) {
      KeywordVO k = (KeywordVO)it.next();

      sqlOut.write("INSERT INTO event_keywords VALUES (");
      sqlInt(evid);
      sqlOut.write(',');
      sqlInt(k.getId());
      sqlOut.write(");");

      sqlOut.write('\n');
    }
  }

  /** ===================================================================
   *                       SQL helpers
   *  =================================================================== */

  private void sqlUserId(UserVO val) throws Exception {
    if (val == null) {
      sqlOut.write("NULL");
    } else {
      sqlInt(val.getId());
    }
  }

  private void sqlInt(int val) throws Exception {
    sqlOut.write(String.valueOf(val));
  }

  private void sqlBool(boolean b) throws Exception {
    sqlOut.write('\'');
    if (b) {
      sqlOut.write('T');
    } else {
      sqlOut.write('F');
    }
    sqlOut.write('\'');
  }

  private void sqlChar(char ch) throws Exception {
    sqlOut.write('\'');
    sqlOut.write(ch);
    sqlOut.write('\'');
  }

  private void sqlAccount(UserVO u) throws Exception {
    if (u == null) {
      sqlOut.write("NULL");
    } else {
      sqlOut.write("'");
      sqlOut.write(u.getAccount());
      sqlOut.write("'");
    }
  }

  private void sqlString(String s) throws Exception {
    if (s == null) {
      sqlOut.write("NULL");
    } else {
      sqlOut.write("'");
      int st = 0;
      do {
        int end = s.indexOf('\'', st);
        if (end < 0) {
          sqlOut.write(s.substring(st));
          break;
        }
        sqlOut.write(s.substring(st, end + 1));
        sqlOut.write("'");
        st = end + 1;
      } while (st < s.length());

      sqlOut.write("'");
    }
  }

  private void sqlTimestamp(Timestamp ts) throws Exception {
    if (ts == null) {
      sqlOut.write("NULL");
    } else {
      String s = ts.toString();

      // remove millis.

      int len = s.lastIndexOf('.');
      sqlString(s.substring(0, len));
    }
  }

  private void sqlDate(Date dt) throws Exception {
    if (dt == null) {
      sqlOut.write("NULL");
    } else {
      String s = dt.toString();
      sqlString(dt.toString());
    }
  }

  private void sqlTime(Time tm) throws Exception {
    if (tm == null) {
      sqlOut.write("NULL");
    } else {
      String s = tm.toString();
      sqlString(tm.toString());
    }
  }
}

